# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'seed.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QGridLayout, QLabel, QLineEdit,
    QPushButton, QSizePolicy, QWidget)
import airunner.resources_light_rc
import airunner.resources_dark_rc

class Ui_seed_widget(object):
    def setupUi(self, seed_widget):
        if not seed_widget.objectName():
            seed_widget.setObjectName(u"seed_widget")
        seed_widget.resize(558, 92)
        self.gridLayout = QGridLayout(seed_widget)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.lineEdit = QLineEdit(seed_widget)
        self.lineEdit.setObjectName(u"lineEdit")
        font = QFont()
        font.setPointSize(8)
        self.lineEdit.setFont(font)

        self.gridLayout.addWidget(self.lineEdit, 1, 0, 1, 1)

        self.random_button = QPushButton(seed_widget)
        self.random_button.setObjectName(u"random_button")
        self.random_button.setMinimumSize(QSize(24, 24))
        self.random_button.setMaximumSize(QSize(24, 24))
        self.random_button.setCursor(QCursor(Qt.PointingHandCursor))
        icon = QIcon()
        icon.addFile(u":/icons/dark/dice-game-icon.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.random_button.setIcon(icon)
        self.random_button.setCheckable(True)
        self.random_button.setFlat(True)

        self.gridLayout.addWidget(self.random_button, 1, 1, 1, 1)

        self.label = QLabel(seed_widget)
        self.label.setObjectName(u"label")
        font1 = QFont()
        font1.setPointSize(8)
        font1.setBold(True)
        self.label.setFont(font1)

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)


        self.retranslateUi(seed_widget)
        self.lineEdit.textChanged.connect(seed_widget.action_value_changed_seed)
        self.random_button.toggled.connect(seed_widget.action_clicked_button_random_seed)

        QMetaObject.connectSlotsByName(seed_widget)
    # setupUi

    def retranslateUi(self, seed_widget):
        seed_widget.setWindowTitle(QCoreApplication.translate("seed_widget", u"Form", None))
        self.lineEdit.setPlaceholderText(QCoreApplication.translate("seed_widget", u"seed", None))
#if QT_CONFIG(tooltip)
        self.random_button.setToolTip(QCoreApplication.translate("seed_widget", u"Toggle random seed", None))
#endif // QT_CONFIG(tooltip)
        self.random_button.setText("")
        self.label.setText(QCoreApplication.translate("seed_widget", u"Seed", None))
    # retranslateUi




from airunner.enums import ImageGenerator, GeneratorSection
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.image_generator_preferences.templates.image_generator_preferences_ui import Ui_image_generator_preferences


class ImageGeneratorPreferencesWidget(BaseWidget):
    widget_class_ = Ui_image_generator_preferences

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.initialize_categories()
        self.initialize_pipeline()
        self.initialize_action()

    def initialize_categories(self):
        self.ui.category.blockSignals(True)
        self.ui.category.clear()
        for item in ImageGenerator:
            self.ui.category.addItem(item.value)
        self.ui.category.setCurrentText(self.application_settings.current_image_generator)
        self.ui.category.blockSignals(False)

    def initialize_pipeline(self):
        self.ui.pipeline.blockSignals(True)
        self.ui.pipeline.clear()
        for item in ImageGenerator:
            self.ui.pipeline.addItem(item.value)
        self.ui.pipeline.setCurrentText(self.application_settings.current_image_generator)
        self.ui.pipeline.blockSignals(False)

    def initialize_action(self):
        self.ui.action.blockSignals(True)
        self.ui.action.clear()
        for item in GeneratorSection:
            self.ui.action.addItem(item.value)
        self.ui.action.setCurrentText(self.application_settings.current_image_generator)
        self.ui.action.blockSignals(False)

    def stablediffusion_toggled(self, val):
        if val:
            self.update_application_settings("current_image_generator", ImageGenerator.STABLEDIFFUSION.value)
            self.update_application_settings("generator_section", GeneratorSection.TXT2IMG.value)

    def category_changed(self, val):
        self.update_application_settings("generator_section", val)

    def pipeline_changed(self, val):
        self.update_application_settings("current_image_generator", val)

    def version_changed(self, val):
        self.update_application_settings("current_image_generator", val)

    def action_changed(self, val):
        self.update_application_settings("current_image_generator", val)

    def model_changed(self, val):
        self.update_application_settings("current_image_generator", val)




import unittest

from airunner.handlers.stablediffusion.prompt_weight_bridge import PromptWeightBridge


class TestPromptWeightConvert(unittest.TestCase):
    def test_simple(self):
        prompt = "Example (a GHI:1.4)"
        expected_prompt = "Example (a GHI)1.4"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

    def test_use_case_a(self):
        prompt = "Example (ABC): 1.23 XYZ (DEF) (GHI:1.4)"
        expected_prompt = "Example (ABC)1.1: 1.23 XYZ (DEF)1.1 (GHI)1.4"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)
    #
    def test_use_case_b(self):
        prompt = "(a dog:0.5) and a cat"
        expected_prompt = "(a dog)0.5 and a cat"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

    def test_use_case_c(self):
        prompt = "A perfect photo of a woman wearing a respirator wandering through the (toxic wasteland:1.3)"
        expected_prompt = "A perfect photo of a woman wearing a respirator wandering through the (toxic wasteland)1.3"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

    def test_use_case_d(self):
        prompt = "(worst quality:0.8), fantasy, cartoon, halftone print, (cinematic:1.2), verybadimagenegative_v1.3, easynegative, (surreal:0.8), (modernism:0.8), (art deco:0.8), (art nouveau:0.8)"
        expected_prompt = "(worst quality)0.8, fantasy, cartoon, halftone print, (cinematic)1.2, verybadimagenegative_v1.3, easynegative, (surreal)0.8, (modernism)0.8, (art deco)0.8, (art nouveau)0.8"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

    def test_convert_basic_parentheses(self):
        prompt = "(a hammer) and a cat in a car"
        expected_prompt = "(a hammer)1.1 and a cat in a car"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "a foo and (a cat) in a car"
        expected_prompt = "a foo and (a cat)1.1 in a car"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "(a bar) and (a cat) in a car"
        expected_prompt = "(a bar)1.1 and (a cat)1.1 in a car"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "(a baz) and (a cat) in (a car)"
        expected_prompt = "(a baz)1.1 and (a cat)1.1 in (a car)1.1"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "((a car)) and a cat"
        expected_prompt = "(a car)1.21 and a cat"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "(((((((((((a car))))))))))) and a cat"
        expected_prompt = "(a car)1.4 and a cat"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "(((((((((((a car))))))))))) and (((a cat)))"
        expected_prompt = "(a car)1.4 and (a cat)1.33"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

    def test_convert_basic_brackets(self):
        prompt = "[a hammer] and a cat in a car"
        expected_prompt = "(a hammer)0.9 and a cat in a car"
        self.assertEqual(PromptWeightBridge.convert_basic_brackets(prompt), expected_prompt)

        prompt = "a foo and [a cat] in a car"
        expected_prompt = "a foo and (a cat)0.9 in a car"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "[a bar] and [a cat] in a car"
        expected_prompt = "(a bar)0.9 and (a cat)0.9 in a car"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "[a baz] and [a cat] in [a car]"
        expected_prompt = "(a baz)0.9 and (a cat)0.9 in (a car)0.9"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "[[a car]] and a cat"
        expected_prompt = "(a car)0.79 and a cat"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "[[[[[[[[[[[a car]]]]]]]]]]] and a cat"
        expected_prompt = "(a car)0.0 and a cat"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "[[[[[[[[[[[a car]]]]]]]]]]] and [[[a cat]]]"
        expected_prompt = "(a car)0.0 and (a cat)0.67"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

    def test_convert_prompt_with_weight_value(self):
        prompt = "(a bird:0.5) and a plane"
        expected_prompt = "(a bird)0.5 and a plane"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "a dog and (a cat:0.6)"
        expected_prompt = "a dog and (a cat)0.6"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "(a boat:0.5) and (a ship:0.6)"
        expected_prompt = "(a boat)0.5 and (a ship)0.6"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "(a man:0.5) and a woman"
        expected_prompt = "(a man)0.5 and a woman"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

    def test_compel_prompt_weight(self):
        prompt = "(a asdf)1.4 and a cat in a car"
        expected_prompt = "(a asdf)1.4 and a cat in a car"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "a man (eating an apple)++++"
        expected_prompt = "a man (eating an apple)++++"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "a man (eating fruit)+"
        expected_prompt = "a man (eating fruit)+"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "a man (eating a bannana)----"
        expected_prompt = "a man (eating a bannana)----"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "a man (eating bread)-"
        expected_prompt = "a man (eating bread)-"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = '("blue sphere", "red cube").blend(0.25,0.75)'
        expected_prompt = prompt
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

    def test_mixed_prompt(self):
        prompt = "a man (drinking (water))0.5"
        expected_prompt = "a man (drinking (water)1.1)0.5"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "a man (drinking ((milk)))0.5"
        expected_prompt = "a man (drinking (milk)1.21)0.5"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "a man (drinking+ ((juice)))0.5"
        expected_prompt = "a man (drinking+ (juice)1.21)0.5"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "a man ((drinking:1.1) ((beer)))0.5"
        expected_prompt = "a man ((drinking)1.1 (beer)1.21)0.5"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "a fox (drinking) water in a [bar] from (the country) of [canada]"
        expected_prompt = "a fox (drinking)1.1 water in a (bar)0.9 from (the country)1.1 of (canada)0.9"
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

    def test_blend_conversion(self):
        prompt = "A [frog:turtle:0.1] on a leaf in the forest"
        expected_prompt = 'A ("frog", "turtle").blend(0.1, 0.9) on a leaf in the forest'
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "A [frog:turtle:0.5] on a leaf in the forest"
        expected_prompt = 'A ("frog", "turtle").blend(0.5, 0.5) on a leaf in the forest'
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)

        prompt = "A (frog:turtle:0.1) on a leaf in the forest"
        expected_prompt = 'A ("frog", "turtle").blend(0.9, 0.1) on a leaf in the forest'
        self.assertEqual(PromptWeightBridge.convert(prompt), expected_prompt)



import inspect
from typing import Callable
from PySide6.QtCore import QObject, Signal as BaseSignal, Slot
from airunner.enums import SignalCode


class SingletonMeta(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class Signal(QObject):
    """
    This class represents a signal that can be emitted and received.
    """
    signal = BaseSignal(dict)

    def __init__(self, callback: Callable):
        super().__init__()
        self.callback = callback

        self.signal.connect(self.on_signal_received)

    @Slot(object)
    def on_signal_received(self, data: dict):
        # Check if the callback expects a parameter
        if len(inspect.signature(self.callback).parameters) == 0:
            self.callback()
        else:
            self.callback(data)


class SignalMediator(metaclass=SingletonMeta):
    """
    This class is responsible for mediating signals between classes.
    """

    signals = {}

    def register(
        self,
        code: SignalCode,
        slot_function: Callable
    ):
        """
        Register a signal to be received by a class.

        :param code: The SignalCode of the signal to register
        :param slot_function: The function to call when the signal is received.
        """
        # Create a new Signal instance for this signal name
        if code not in self.signals:
            self.signals[code] = []
        self.signals[code].append(Signal(callback=slot_function))

    def emit_signal(
        self,
        code: SignalCode,
        data: object = None
    ):
        """
        Emit a signal.
        :param code:
        :param data:
        :return:
        """
        data = {} if data is None else data
        if code in self.signals:
            for signal in self.signals[code]:
                signal.signal.emit(data)



import os

from PySide6 import QtGui
from PySide6.QtWidgets import QWidget

from airunner.enums import CanvasToolName
from airunner.windows.main.settings_mixin import SettingsMixin
from airunner.mediator_mixin import MediatorMixin
from airunner.utils.create_worker import create_worker


class BaseWidget(
    QWidget,
    MediatorMixin,
    SettingsMixin
):
    widget_class_ = None
    icons = ()
    ui = None
    qss_filename = None
    threads = []

    @property
    def current_tool(self):
        return CanvasToolName(self.application_settings.current_tool)

    @property
    def is_dark(self):
        return self.application_settings.dark_mode_enabled

    def __init__(self, *args, **kwargs):
        MediatorMixin.__init__(self)
        
        super().__init__(*args, **kwargs)
        if self.widget_class_:
            self.ui = self.widget_class_()
        if self.ui:
            self.ui.setupUi(self)
            self.set_icons()

        self.signal_handlers: dict = {}
        self.services: dict = {}
        self.worker_class_map: dict = {}

    def initialize(self):
        """
        Call this function to initialize the widget.
        :return:
        """
        self.register_signals()
        self.initialize_workers()
        self.initialize_form()

    def register_signals(self):
        """
        Set signal_handlers dict in order to register signals.

        signal_handlers should be a dictionary of SignalCode enums and functions.
        Example:
        signal_handlers = {
            SignalCode.GET_SETTINGS: self.get_settings,
            SignalCode.SET_SETTINGS: self.set_settings
        }
        :return:
        """
        for signal, handler in self.signal_handlers.items():
            self.register(signal, handler)

    def initialize_workers(self):
        """
        Override this function to initialize workers.

        worker_class_map should be a dictionary of property names and worker classes.
        Example:
        worker_class_map = {
            "worker": WorkerClass
        }
        :return:
        """
        for property_name, worker_class_name_ in self.worker_class_map.items():
            worker = create_worker(worker_class_name_)
            setattr(self, property_name, worker)

    def initialize_form(self):
        pass

    def add_to_grid(self, widget, row, column, row_span=1, column_span=1):
        self.layout().addWidget(widget, row, column, row_span, column_span)
    
    def showEvent(self, event):
        super().showEvent(event)
        """
        Triggered when the app is loaded.
        Override this function in order to initialize the widget rather than
        using __init__.
        """
        self.initialize()

    def set_icons(self):
        theme = "dark" if self.is_dark else "light"
        for icon_data in self.icons:
            icon_name = icon_data[0]
            widget_name = icon_data[1]
            icon = QtGui.QIcon()
            icon.addPixmap(
                QtGui.QPixmap(f":/icons/{theme}/{icon_name}.svg"),
                QtGui.QIcon.Mode.Normal,
                QtGui.QIcon.State.Off)
            getattr(self.ui, widget_name).setIcon(icon)
        self.update()

    def set_button_icon(self, is_dark, button_name, icon):
        try:
            getattr(self, button_name).setIcon(
                QtGui.QIcon(
                    os.path.join(f"src/icons/{icon}{'-light' if is_dark else ''}.png")
                )
            )
        except AttributeError as e:
            pass

    def get_form_element(self, element):
        return getattr(self.ui, element)

    def get_plain_text(self, element):
        try:
            return self.get_form_element(element).toPlainText()
        except AttributeError:
            return None

    def get_text(self, element):
        try:
            return self.get_form_element(element).text()
        except AttributeError:
            return None

    def get_value(self, element):
        try:
            return self.get_form_element(element).value()
        except AttributeError:
            return None

    def get_is_checked(self, element):
        try:
            return self.get_form_element(element).isChecked()
        except AttributeError:
            return None

    def set_plain_text(self, element, val):
        try:
            self.get_form_element(element).setPlainText(val)
            return True
        except AttributeError:
            return False

    def set_text(self, element, val):
        try:
            self.get_form_element(element).setText(val)
            return True
        except AttributeError:
            return False
        except TypeError:
            return False

    def set_value(self, element, val):
        try:
            self.get_form_element(element).setValue(val)
            return True
        except AttributeError:
            return False

    def set_is_checked(self, element, val):
        try:
            self.get_form_element(element).setChecked(val)
            return True
        except AttributeError:
            return False



import datetime
import json
import os
import sqlite3
import time
import traceback
from typing import AnyStr
import torch
from PySide6.QtCore import QObject

from llama_index.core import Settings, RAKEKeywordTableIndex
from llama_index.core.base.llms.types import ChatMessage
from llama_index.readers.file import EpubReader, PDFReader, MarkdownReader
from llama_index.core import SimpleDirectoryReader
from llama_index.core.node_parser import SentenceSplitter
from llama_index.core import PromptHelper
from llama_index.core.chat_engine import ContextChatEngine
from llama_index.core.indices.keyword_table import KeywordTableSimpleRetriever
from transformers import TextIteratorStreamer

from airunner.data.models.settings_models import Conversation
from airunner.handlers.llm.huggingface_llm import HuggingFaceLLM
from airunner.handlers.llm.custom_embedding import CustomEmbedding
from airunner.handlers.llm.agent.html_file_reader import HtmlFileReader
from airunner.handlers.llm.agent.external_condition_stopping_criteria import ExternalConditionStoppingCriteria
from airunner.mediator_mixin import MediatorMixin
from airunner.enums import (
    SignalCode,
    LLMChatRole,
    LLMActionType,
    AgentState
)
from airunner.utils.get_torch_device import get_torch_device
from airunner.utils.create_worker import create_worker
from airunner.utils.prepare_llm_generate_kwargs import prepare_llm_generate_kwargs
from airunner.windows.main.settings_mixin import SettingsMixin
from airunner.workers.agent_worker import AgentWorker


class RefreshContextChatEngine(ContextChatEngine):
    def stream_chat(self, *args, system_prompt:str=None, **kwargs):
        if system_prompt:
            self._prefix_messages[0] = ChatMessage(content=system_prompt, role=self._llm.metadata.system_role)
        return super().stream_chat(*args, **kwargs)


class BaseAgent(
    QObject,
    MediatorMixin,
    SettingsMixin
):
    def __init__(self, *args, **kwargs):
        MediatorMixin.__init__(self)

        self._requested_action = None
        self.model = kwargs.pop("model", None)
        self.__documents = None
        self.__document_reader: SimpleDirectoryReader = None
        self.__index = None
        self.__chat_engine = None
        self.__retriever = None
        self.__storage_context = None
        self.__transformations = None
        self.__index_struct = None
        self.__callback_manager = None
        self.__pdf_reader = None
        self.__epub_reader = None
        self.__html_reader = None
        self.__markdown_reader = None
        self.__embedding = None
        self.__model_name = os.path.expanduser(
            os.path.join(
                self.path_settings.base_path,
                "text/models",
                "sentence_transformers/sentence-t5-large"
            )
        )
        self.__query_instruction = "Search through all available texts and provide a brief summary of the key points which are relevant to the query."
        self.__text_instruction = "Summarize and provide a brief explanation of the text. Stay concise and to the point."
        self.__state = AgentState.SEARCH
        self.__chunk_size = 1000
        self.__chunk_overlap = 512
        self.__target_files = []
        self.__rag_model = None
        self.__rag_tokenizer = None
        self.__model = None
        self.__tokenizer = None
        self.__llm = None
        self._chatbot = None
        self.action = LLMActionType.CHAT
        self.rendered_template = None
        self.tokenizer = kwargs.pop("tokenizer", None)
        self._streamer = None
        self.chat_template = kwargs.pop("chat_template", "")
        self.is_mistral = kwargs.pop("is_mistral", True)
        self.conversation_id = None
        self.conversation_title = None
        self.conversation = None
        self.history = self.load_history_from_db(self.conversation_id)  # Load history by conversation ID
        super().__init__(*args, **kwargs)
        self.prompt = ""
        self.thread = None
        self._do_interrupt = False
        self.response_worker = create_worker(AgentWorker)
        self.load_rag(model=self.model, tokenizer=self.tokenizer)

    @property
    def do_interrupt(self):
        return self._do_interrupt

    @do_interrupt.setter
    def do_interrupt(self, value):
        self._do_interrupt = value

    @property
    def streamer(self):
        if self._streamer is None:
            self._streamer = TextIteratorStreamer(self.tokenizer)
        return self._streamer

    @property
    def available_actions(self):
        return {
            0: LLMActionType.TOGGLE_FULLSCREEN,
            1: LLMActionType.TOGGLE_TTS,
            2: LLMActionType.GENERATE_IMAGE,
            3: LLMActionType.PERFORM_RAG_SEARCH,
            4: LLMActionType.CHAT,
        }

    @property
    def username(self) -> str:
        return self.user.username

    @property
    def botname(self) -> str:
        return self.chatbot.botname

    @property
    def bot_mood(self) -> str:
        return self.conversation.bot_mood if self.conversation else ""

    @bot_mood.setter
    def bot_mood(self, value: str):
        conversation = self.conversation
        conversation.bot_mood = value
        self.save_object(conversation)
        self.emit_signal(SignalCode.BOT_MOOD_UPDATED, {
            "mood": value
        })

    @property
    def bot_personality(self) -> str:
        return self.chatbot.bot_personality

    @property
    def override_parameters(self):
        generate_kwargs = prepare_llm_generate_kwargs(self.llm_generator_settings)
        return generate_kwargs if self.llm_generator_settings.override_parameters else {}

    @property
    def system_instructions(self):
        return self.chatbot.system_instructions

    @property
    def generator_settings(self) -> dict:
        return prepare_llm_generate_kwargs(self.chatbot)

    @property
    def device(self):
        return get_torch_device(self.memory_settings.default_gpu_llm)

    @property
    def target_files(self):
        return [
            target_file.file_path for target_file in self.chatbot.target_files
        ]

    @property
    def query_instruction(self):
        if self.__state == AgentState.SEARCH:
            return self.__query_instruction
        elif self.__state == AgentState.CHAT:
            return "Search through the chat history for anything relevant to the query."

    @property
    def text_instruction(self):
        if self.__state == AgentState.SEARCH:
            return self.__text_instruction
        elif self.__state == AgentState.CHAT:
            return "Use the text to respond to the user"

    @property
    def index(self):
        if self.__state == AgentState.SEARCH:
            return self.__index
        elif self.__state == AgentState.CHAT:
            return self.__chat_history_index

    @property
    def llm(self):
        if self.__llm is None:
            try:
                if self.llm_generator_settings.use_api:
                    self.__llm = self.__model
                else:
                    self.__llm = HuggingFaceLLM(model=self.__model, tokenizer=self.__tokenizer)
            except Exception as e:
                self.logger.error(f"Error loading LLM: {str(e)}")
        return self.__llm

    @property
    def chat_engine(self):
        return self.__chat_engine

    @property
    def is_llama_instruct(self):
        return True

    @property
    def use_cuda(self):
        return torch.cuda.is_available()

    @property
    def cuda_index(self):
        return 0

    def unload(self):
        self.unload_rag()
        del self.model
        del self.tokenizer
        self.model = None
        self.tokenizer = None
        self.thread = None

    def clear_history(self):
        self.history = []
        self.reload_rag()
        self.conversation_id = None
        self.conversation_title = None

    def _update_conversation_title(self, title):
        self.conversation_title = title
        self.update_conversation_title(self.conversation_id, title)

    def _create_conversation(self):
        # Get the most recent conversation ID
        self.conversation = self.get_most_recent_conversation()
        if not self.conversation:
            self.conversation = self.create_conversation()
        self.conversation_id = self.conversation.id
        self.history = self.load_history_from_db(self.conversation_id)

    def interrupt_process(self):
        self.do_interrupt = True

    def do_interrupt_process(self):
        if self.do_interrupt:
            self.emit_signal(
                SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                dict(
                    message="",
                    is_first_message=False,
                    is_end_of_message=False,
                    name=self.botname,
                    action=LLMActionType.CHAT
                )
            )
        return self.do_interrupt

    def mood(self, botname: str, bot_mood: str, use_mood: bool) -> str:
        return (
            f"{botname} is not a computer or a bot, you are a human with emotions. "
            f"{botname}'s mood changes depending on the conversation and other factors.\n"
            f"{botname}'s current mood: {bot_mood}."
            #f"{botname}'s evaluation of {username}: {self.user_evaluation}"
        ) if (use_mood and bot_mood != "") else ""

    def history_prompt(self) -> str:
        history = ""
        previous_date = None
        for message in self.history:
            # Check if the timestamp is a string and convert it to a datetime object if necessary
            if isinstance(message["timestamp"], str):
                current_date = datetime.datetime.strptime(message["timestamp"], "%Y-%m-%d %H:%M:%S").date()
            else:
                current_date = message["timestamp"].date()

            if previous_date is None or current_date != previous_date:
                history += f"\n--- {current_date.strftime('%B %d, %Y')} ---\n"
                previous_date = current_date
            name = ""
            if message["role"] == LLMChatRole.HUMAN.value:
                name = self.username
            elif message["role"] == LLMChatRole.ASSISTANT.value:
                name = self.botname
            history += f"{name}: {message['content']}\n"
        return (
            "------\n"
            "Chat History:\n"
            f"{history}"
        )

    def date_time_prompt(self) -> str:
        if not self.chatbot.use_datetime:
            return ""
        current_date = datetime.datetime.now().strftime("%A %b %d, %Y")
        current_time = datetime.datetime.now().strftime("%I:%M:%S %p")
        current_timezone = time.tzname
        prompt = [
            "\n======\n",
            f"Use the following information to help you with your response, but do not include it in your response or reference it directly unless asked.",
            f"Date: {current_date}, Time: {current_time}, Timezone: {current_timezone}",
        ]
        return "\n".join(prompt)

    def build_system_prompt(
        self,
        action
    ):
        system_instructions = ""
        guardrails_prompt = ""
        use_mood = self.chatbot.use_mood
        use_personality = self.chatbot.use_personality
        use_names = self.chatbot.assign_names
        use_system_instructions = self.chatbot.use_system_instructions
        use_guardrails = self.chatbot.use_guardrails
        bot_mood = self.bot_mood
        bot_personality = self.chatbot.bot_personality
        username = self.user.username
        botname = self.chatbot.botname
        if use_system_instructions:
            system_instructions = self.chatbot.system_instructions
        if use_guardrails:
            guardrails_prompt = self.chatbot.guardrails_prompt

        system_prompt = []

        if action is LLMActionType.CHAT:
            """
            Build the system prompt for chat template
            """
            system_prompt = [
                guardrails_prompt,
                system_instructions,
                self.names_prompt(use_names, botname, username),
                self.mood(botname, bot_mood, use_mood),
                self.personality_prompt(bot_personality, use_personality),
                self.history_prompt(),
                self.date_time_prompt()
            ]

        elif action is LLMActionType.GENERATE_IMAGE:
            system_prompt = [
                system_instructions,
                self.history_prompt()
            ]

        elif action is LLMActionType.APPLICATION_COMMAND:
            prompt_template = self.get_prompt_template_by_name("application_command")
            system_instructions = prompt_template.system

            # Create a list of commands that the bot can choose from
            for index, action in self.available_actions.items():
                system_instructions += f"{index}: {action.value}\n"

            system_prompt = [
                self.names_prompt(use_names, botname, username),
                self.mood(botname, bot_mood, use_mood),
                system_instructions,
                self.history_prompt(),
            ]

        elif action is LLMActionType.SUMMARIZE:
            prompt_template = self.get_prompt_template_by_name("summarize")
            system_instructions = prompt_template.system
            system_prompt = [
                system_instructions,
                self.history_prompt()
            ]

        elif action is LLMActionType.UPDATE_MOOD:
            prompt_template = self.get_prompt_template_by_name("update_mood")
            system_instructions = prompt_template.system
            system_prompt = [
                system_instructions,
                self.names_prompt(use_names, botname, username),
                self.mood(botname, bot_mood, use_mood),
                self.personality_prompt(bot_personality, use_personality),
                self.history_prompt(),
            ]

        elif action is LLMActionType.PERFORM_RAG_SEARCH:
            prompt_template = self.get_prompt_template_by_name("rag_search")
            system_instructions = prompt_template.system
            system_prompt = [
                guardrails_prompt,
                system_instructions,
                self.names_prompt(use_names, botname, username),
                self.mood(botname, bot_mood, use_mood),
                self.personality_prompt(bot_personality, use_personality),
                # self.history_prompt(),
            ]

        elif action is LLMActionType.QUIT_APPLICATION:
            self.emit_signal(SignalCode.QUIT_APPLICATION)

        elif action is LLMActionType.TOGGLE_FULLSCREEN:
            self.emit_signal(SignalCode.TOGGLE_FULLSCREEN_SIGNAL)

        return "\n".join(system_prompt)

    def names_prompt(self, use_names: bool, botname: str, username: str) -> str:
        return f"Your name is {botname}. \nThe user's name is {username}." if use_names else ""

    def personality_prompt(self, bot_personality: str, use_personality: bool) -> str:
        return (
            f"Your personality: {bot_personality}."
        ) if use_personality else ""

    def prepare_messages(
        self,
        action
    ) -> list:
        system_prompt = self.build_system_prompt(action)
        if action is LLMActionType.APPLICATION_COMMAND:
            prompt = (
                "Choose an action from THE LIST of commands for the text above. "
                "Only return the number of the command."
            )
        elif action is LLMActionType.GENERATE_IMAGE:
            prompt = (
                f"Replace the placeholder values in the following JSON:\n"
                "```json\n"+ json.dumps(dict(
                    description="PLACEHOLDER",
                    composition="PLACEHOLDER"
                )) +"\n```\n"
            )
        elif action is LLMActionType.SUMMARIZE:
            prompt = (
                f"Summarize the conversation history"
            )
        else:
            prompt = f"Respond to {self.username}"
        messages = [
            {
                "content": system_prompt,
                "role": LLMChatRole.SYSTEM.value
            },
            {
                "content": prompt,
                "role": LLMChatRole.HUMAN.value
            }
        ]
        return messages

    def get_rendered_template(
        self,
        action: LLMActionType
    ) -> str:
        conversation = self.prepare_messages(action)
        rendered_template = self.tokenizer.apply_chat_template(
            chat_template=self.chat_template,
            conversation=conversation,
            tokenize=False
        )

        # replace variables in chat template
        variables = {
            "speaker_name": self.botname,
            "listener_name": self.username,
            "username": self.username,
            "botname": self.botname,
            "bot_mood": self.bot_mood,
            "bot_personality": self.bot_personality,
        }
        for key, value in variables.items():
            value = value or ""
            rendered_template = rendered_template.replace("{{ " + key + " }}", value)
        return rendered_template

    def run(
        self,
        prompt: str,
        action: LLMActionType,
        **kwargs
    ):
        self._requested_action = None
        self.action = action

        if action is LLMActionType.TOGGLE_TTS:
            self.emit_signal(SignalCode.TOGGLE_TTS_SIGNAL)
            return

        if action in (
            LLMActionType.CHAT,
            LLMActionType.PERFORM_RAG_SEARCH
        ) and (self.conversation_title is None or self.conversation_title == ""):
            self._requested_action = action
            action = LLMActionType.SUMMARIZE

        self.logger.debug("Running...")
        self.prompt = prompt

        if self.conversation_id is None:
            self._create_conversation()
            self.set_conversation_title()

        # Add the user's message to history
        if action not in (
            LLMActionType.APPLICATION_COMMAND,
            LLMActionType.UPDATE_MOOD
        ):
            self.add_message_to_history(
                self.prompt,
                LLMChatRole.HUMAN
            )

        self.rendered_template = self.get_rendered_template(action)

        model_inputs = self.tokenizer(
            self.rendered_template,
            return_tensors="pt"
        ).to(self.device)

        kwargs.update(
            action=action,
            do_emit_response=True
        )

        self.run_with_thread(
            model_inputs,
            **kwargs
        )

    def prepare_generate_data(self, model_inputs, stopping_criteria):
        data = dict(
            **model_inputs,
            **self.generator_settings,
            stopping_criteria=[stopping_criteria]
        )
        data.update(self.override_parameters)
        return data

    def run_with_thread(
        self,
        model_inputs,
        action: LLMActionType,
        **kwargs,
    ):
        with torch.no_grad():
            # Generate the response
            self.logger.debug("Generating...")

            self.emit_signal(SignalCode.UNBLOCK_TTS_GENERATOR_SIGNAL)

            if self.do_interrupt:
                self.do_interrupt = False

            stopping_criteria = ExternalConditionStoppingCriteria(self.do_interrupt_process)
            data = self.prepare_generate_data(model_inputs, stopping_criteria)

            data["streamer"] = kwargs.get("streamer", self.streamer)

            if action is not LLMActionType.PERFORM_RAG_SEARCH:
                try:
                    self.response_worker.add_to_queue({
                        "model": self.model,
                        "kwargs": data,
                        "prompt": self.prompt,
                        "botname": self.botname,
                    })
                except Exception as e:
                    self.logger.error("545: An error occurred in model.generate:")
                    self.logger.error(str(e))
                    self.logger.error(traceback.format_exc())

            # strip all new lines from rendered_template:
            #self.rendered_template = self.rendered_template.replace("\n", " ")
            eos_token = self.tokenizer.eos_token
            bos_token = self.tokenizer.bos_token
            if self.is_mistral:
                self.rendered_template = bos_token + self.rendered_template
            skip = True
            streamed_template = ""
            replaced = False
            is_end_of_message = False
            is_first_message = True

            if action == LLMActionType.GENERATE_IMAGE:
                self.emit_signal(
                    SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                    dict(
                        message="Generating image prompt.\n",
                        is_first_message=is_first_message,
                        is_end_of_message=False,
                        name=self.botname,
                        action=LLMActionType.CHAT
                    )
                )
                is_first_message = False

            if action in (
                LLMActionType.CHAT,
                LLMActionType.GENERATE_IMAGE,
                LLMActionType.UPDATE_MOOD,
                LLMActionType.SUMMARIZE,
                LLMActionType.APPLICATION_COMMAND
            ):
                for new_text in self.streamer:
                    # strip all newlines from new_text
                    streamed_template += new_text
                    if self.is_mistral:
                        streamed_template = streamed_template.replace(f"{bos_token} [INST]", f"{bos_token}[INST]")
                        streamed_template = streamed_template.replace("  [INST]", " [INST]")
                    # iterate over every character in rendered_template and
                    # check if we have the same character in streamed_template
                    if not replaced:
                        for i, char in enumerate(self.rendered_template):
                            try:
                                if char == streamed_template[i]:
                                    skip = False
                                else:
                                    skip = True
                                    break
                            except IndexError:
                                skip = True
                                break
                    if skip:
                        continue
                    elif not replaced:
                        replaced = True
                        streamed_template = streamed_template.replace(self.rendered_template, "")
                    else:
                        if eos_token in new_text:
                            streamed_template = streamed_template.replace(eos_token, "")
                            new_text = new_text.replace(eos_token, "")
                            is_end_of_message = True
                        # strip botname from new_text
                        new_text = new_text.replace(f"{self.botname}:", "")
                        if action in (
                            LLMActionType.CHAT,
                            LLMActionType.PERFORM_RAG_SEARCH,
                            LLMActionType.GENERATE_IMAGE
                        ):
                            self.emit_signal(
                                SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                                dict(
                                    message=new_text,
                                    is_first_message=is_first_message,
                                    is_end_of_message=is_end_of_message,
                                    name=self.botname,
                                    action=action
                                )
                            )
                        else:
                            self.emit_signal(
                                SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                                dict(
                                    message="",
                                    is_first_message=is_first_message,
                                    is_end_of_message=is_end_of_message,
                                    name=self.botname,
                                    action=action
                                )
                            )
                        is_first_message = False
            elif action is LLMActionType.PERFORM_RAG_SEARCH:
                streamed_template = ""
                data = dict(
                    **self.generator_settings,
                    stopping_criteria=[stopping_criteria]
                )
                data.update(self.override_parameters)
                self.llm.generate_kwargs = data
                response = self.chat_engine.stream_chat(
                    message=self.prompt,
                    system_prompt=self.rendered_template
                )
                is_first_message = True
                is_end_of_message = False
                for new_text in response.response_gen:
                    streamed_template += new_text

                    self.emit_signal(
                        SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                        dict(
                            message=" " +new_text,
                            is_first_message=is_first_message,
                            is_end_of_message=is_end_of_message,
                            name=self.botname,
                            action=action
                        )
                    )
                    is_first_message = False
                self.emit_signal(
                    SignalCode.LLM_TEXT_STREAMED_SIGNAL,
                    dict(
                        message="",
                        is_first_message=False,
                        is_end_of_message=True,
                        name=self.botname,
                        action=action
                    )
                )

            if streamed_template is not None:
                if action in (
                    LLMActionType.CHAT,
                    LLMActionType.PERFORM_RAG_SEARCH,
                ):
                    self.add_message_to_history(
                        streamed_template,
                        LLMChatRole.ASSISTANT
                    )

                elif action is LLMActionType.UPDATE_MOOD:
                    self.bot_mood = streamed_template
                    return self.run(
                        prompt=self.prompt,
                        action=LLMActionType.CHAT,
                        **kwargs,
                    )

                elif action is LLMActionType.SUMMARIZE:
                    self._update_conversation_title(streamed_template)
                    return self.run(
                        prompt=self.prompt,
                        action=self._requested_action,
                        **kwargs,
                    )

                elif action is LLMActionType.GENERATE_IMAGE:
                    self.emit_signal(
                        SignalCode.LLM_IMAGE_PROMPT_GENERATED_SIGNAL,
                        {
                            "message": streamed_template,
                            "type": "photo"
                        }
                    )
                elif action is LLMActionType.APPLICATION_COMMAND:
                    index = ''.join(c for c in streamed_template if c.isdigit())
                    try:
                        index = int(index)
                    except ValueError:
                        index = 0

                    try:
                        action = self.available_actions[index]
                    except KeyError:
                        action = LLMActionType.CHAT

                    if action is not None:
                        return self.run(
                            prompt=self.prompt,
                            action=action,
                        )
            return streamed_template

    def get_db_connection(self):
        return sqlite3.connect('airunner.db')

    def add_message_to_history(
        self,
        content: AnyStr,
        role: LLMChatRole = LLMChatRole.ASSISTANT
    ):
        if content is None:
            return

        name = self.username
        is_bot = False

        if role is LLMChatRole.ASSISTANT and content:
            content = content.replace(f"{self.botname}:", "")
            content = content.replace(f"{self.botname}", "")
            is_bot = True
            name = self.botname

        message = self.save_message(
            content,
            role.value,
            name,
            is_bot,
            self.conversation_id
        )

        self.history.append({
            "role": message.role,
            "content": message.content,
            "name": name,
            "is_bot": message.is_bot,
            "timestamp": message.timestamp,
            "conversation_id": message.conversation_id
        })

    def on_load_conversation(self, message):
        self.history = []
        self.conversation = message["conversation"]
        self.conversation_id = message["conversation_id"]

        # Merge the conversation object into the current session
        self.conversation = self.session.merge(self.conversation)

        self.history = self.load_history_from_db(self.conversation_id)
        self.set_conversation_title()
        self.emit_signal(SignalCode.SET_CONVERSATION, {
            "messages": self.history
        })

    def set_conversation_title(self):
        
        self.conversation_title = self.session.query(Conversation).filter_by(id=self.conversation_id).first().title
        

    def load_rag(self, model, tokenizer):
        self.__model = model
        self.__tokenizer = tokenizer
        self.__load_rag()

    def unload_rag(self):
        if self.__llm is not None:
            self.__llm.unload()
            del self.__llm
            self.__llm = None

    def reload_rag(self, data: dict = None):
        self.logger.debug("Reloading RAG index...")
        self.__target_files = data["target_files"] if data is not None else self.__target_files
        self.__load_rag()

    def __load_rag(self):
        self.__load_rag_tokenizer()
        self.__load_rag_model()
        self.__load_embeddings()
        self.__load_readers()
        self.__load_file_extractor()
        self.__load_document_reader()
        self.__load_documents()
        self.__load_text_splitter()
        self.__load_prompt_helper()
        self.__load_settings()
        self.__load_document_index()
        self.__load_retriever()
        self.__load_context_chat_engine()

    def __load_rag_tokenizer(self):
        self.logger.debug("Loading RAG tokenizer...")
        # TODO
        pass

    def __load_rag_model(self):
        self.logger.debug("Loading RAG model...")
        # TODO
        pass

    def __load_embeddings(self):
        self.logger.debug("Loading embeddings...")
        self.__embedding = CustomEmbedding(self.llm)

    def __load_readers(self):
        self.__pdf_reader = PDFReader()
        self.__epub_reader = EpubReader()
        self.__html_reader = HtmlFileReader()
        self.__markdown_reader = MarkdownReader()

    def __load_file_extractor(self):
        self.file_extractor = {
            ".pdf": self.__pdf_reader,
            ".epub": self.__epub_reader,
            ".html": self.__html_reader,
            ".htm": self.__html_reader,
            ".md": self.__markdown_reader,
        }

    def __load_document_reader(self):
        if self.target_files is None or len(self.target_files) == 0:
            return
        self.logger.debug("Loading document reader...")
        try:
            self.__document_reader = SimpleDirectoryReader(
                input_files=self.target_files,
                file_extractor=self.file_extractor,
                exclude_hidden=False
            )
            self.logger.debug("Document reader loaded successfully.")
        except ValueError as e:
            self.logger.error(f"Error loading document reader: {str(e)}")

    def __load_documents(self):
        if not self.__document_reader:
            return
        self.logger.debug("Loading documents...")
        self.__documents = self.__document_reader.load_data()

    def __load_text_splitter(self):
        self.__text_splitter = SentenceSplitter(
            chunk_size=256,
            chunk_overlap=20
        )

    def __load_prompt_helper(self):
        self.__prompt_helper = PromptHelper(
            context_window=4096,
            num_output=1024,
            chunk_overlap_ratio=0.1,
            chunk_size_limit=None,
        )

    def __load_context_chat_engine(self):
        try:
            self.__chat_engine = RefreshContextChatEngine.from_defaults(
                retriever=self.__retriever,
                chat_history=self.history,
                memory=None,
                system_prompt="Search the full text and find all relevant information related to the query.",
                node_postprocessors=[],
                llm=self.__llm,
            )
        except Exception as e:
            self.logger.error(f"Error loading chat engine: {str(e)}")

    def __load_settings(self):
        Settings.llm = self.__llm
        Settings._embed_model = self.__embedding
        Settings.node_parser = self.__text_splitter
        Settings.num_output = 512
        Settings.context_window = 3900

    def __load_document_index(self):
        self.logger.debug("Loading index...")
        documents = self.__documents or []
        try:
            self.__index = RAKEKeywordTableIndex.from_documents(
                documents,
                llm=self.__llm
            )
            self.logger.debug("Index loaded successfully.")
        except TypeError as e:
            self.logger.error(f"Error loading index: {str(e)}")

    def __load_retriever(self):
        try:
            self.__retriever = KeywordTableSimpleRetriever(
                index=self.__index,
            )
            self.logger.debug("Retriever loaded successfully with index.")
        except Exception as e:
            self.logger.error(f"Error setting up the retriever: {str(e)}")



import sounddevice as sd

from queue import Queue

from PySide6.QtCore import QThread

from airunner.enums import SignalCode
from airunner.settings import SLEEP_TIME_IN_MS
from airunner.workers.worker import Worker


class TTSVocalizerWorker(Worker):
    """
    Speech (in the form of numpy arrays generated with the TTS class) is added to the
    vocalizer's queue. The vocalizer plays the speech using sounddevice.
    """
    reader_mode_active = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, signals = (
            (SignalCode.INTERRUPT_PROCESS_SIGNAL, self.on_interrupt_process_signal),
            (SignalCode.UNBLOCK_TTS_GENERATOR_SIGNAL, self.on_unblock_tts_generator_signal),
            (SignalCode.TTS_GENERATOR_WORKER_ADD_TO_STREAM_SIGNAL, self.on_TTSGeneratorWorker_add_to_stream_signal),
        ), **kwargs)
        self.queue = Queue()
        # check if speakers are available
        self.stream = None
        self.start_stream()
        self.started = False
        self.do_interrupt = False
        self.accept_message = True

    def on_interrupt_process_signal(self):
        self.stream.abort()
        self.accept_message = False
        self.queue = Queue()

    def on_unblock_tts_generator_signal(self):
        if self.application_settings.tts_enabled:
            self.logger.debug("Starting TTS stream...")
            self.accept_message = True
            self.stream.start()

    def start_stream(self):
        if sd.query_devices(kind='output'):
            self.stream = sd.OutputStream(samplerate=24000, channels=1)
            self.stream.start()

    def on_TTSGeneratorWorker_add_to_stream_signal(self, response: dict):
        if self.accept_message:
            self.logger.debug("Adding speech to stream...")
            self.add_to_queue(response["message"])

    def handle_message(self, item):
        if not self.accept_message:
            return

        if item is None:
            self.logger.warning("item is none")
            return

        if self.stream is None:
            self.start_stream()
            self.logger.warning("No speakers available")
            return

        # Write the item to the stream
        if self.stream is not None and self.stream.active:
            try:
                self.stream.write(item)
            except sd.PortAudioError:
                self.logger.debug("PortAudioError")
            except AttributeError:
                self.logger.debug("stream is None")

            self.started = True
        QThread.msleep(SLEEP_TIME_IN_MS)

    def handle_speech(self, generated_speech):
        self.logger.debug("Adding speech to stream...")
        try:
            self.queue.put(generated_speech)
        except Exception as e:
            self.logger.error(f"Error while adding speech to stream: {e}")



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'tts_welcome_screen.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QFrame, QGridLayout, QLabel,
    QRadioButton, QSizePolicy, QSpacerItem, QWidget)

class Ui_tts_welcome_screen(object):
    def setupUi(self, tts_welcome_screen):
        if not tts_welcome_screen.objectName():
            tts_welcome_screen.setObjectName(u"tts_welcome_screen")
        tts_welcome_screen.resize(403, 206)
        self.gridLayout = QGridLayout(tts_welcome_screen)
        self.gridLayout.setObjectName(u"gridLayout")
        self.label = QLabel(tts_welcome_screen)
        self.label.setObjectName(u"label")
        font = QFont()
        font.setPointSize(11)
        font.setBold(True)
        self.label.setFont(font)

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.radioButton = QRadioButton(tts_welcome_screen)
        self.radioButton.setObjectName(u"radioButton")
        self.radioButton.setChecked(True)

        self.gridLayout.addWidget(self.radioButton, 4, 0, 1, 1)

        self.radioButton_2 = QRadioButton(tts_welcome_screen)
        self.radioButton_2.setObjectName(u"radioButton_2")

        self.gridLayout.addWidget(self.radioButton_2, 5, 0, 1, 1)

        self.label_2 = QLabel(tts_welcome_screen)
        self.label_2.setObjectName(u"label_2")
        font1 = QFont()
        font1.setBold(True)
        self.label_2.setFont(font1)

        self.gridLayout.addWidget(self.label_2, 3, 0, 1, 1)

        self.verticalSpacer = QSpacerItem(20, 139, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 6, 0, 1, 1)

        self.line = QFrame(tts_welcome_screen)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.Shape.HLine)
        self.line.setFrameShadow(QFrame.Shadow.Sunken)

        self.gridLayout.addWidget(self.line, 2, 0, 1, 1)

        self.label_3 = QLabel(tts_welcome_screen)
        self.label_3.setObjectName(u"label_3")

        self.gridLayout.addWidget(self.label_3, 1, 0, 1, 1)


        self.retranslateUi(tts_welcome_screen)
        self.radioButton.toggled.connect(tts_welcome_screen.yes_toggled)
        self.radioButton_2.toggled.connect(tts_welcome_screen.no_toggled)

        QMetaObject.connectSlotsByName(tts_welcome_screen)
    # setupUi

    def retranslateUi(self, tts_welcome_screen):
        tts_welcome_screen.setWindowTitle(QCoreApplication.translate("tts_welcome_screen", u"Form", None))
        self.label.setText(QCoreApplication.translate("tts_welcome_screen", u"Text-to-Speech setup", None))
        self.radioButton.setText(QCoreApplication.translate("tts_welcome_screen", u"Yes", None))
        self.radioButton_2.setText(QCoreApplication.translate("tts_welcome_screen", u"No", None))
        self.label_2.setText(QCoreApplication.translate("tts_welcome_screen", u"Would you like to use the Text-to-Speech feature?", None))
        self.label_3.setText(QCoreApplication.translate("tts_welcome_screen", u"Text-to-speech requires the download of an AI model.", None))
    # retranslateUi




from airunner.windows.setup_wizard.base_wizard import BaseWizard
from airunner.windows.setup_wizard.model_setup.stable_diffusion_setup.templates.choose_style_ui import Ui_choose_model_style


class ChooseModelStyle(BaseWizard):
    class_name_ = Ui_choose_model_style





controlnet_bootstrap_data = [
    {"display_name": "Canny", "name": "canny", "path": "lllyasviel/control_v11p_sd15_canny", "version": "SD 1.5"},
    {"display_name": "Depth Leres", "name": "depth_leres", "path": "lllyasviel/control_v11f1p_sd15_depth", "version": "SD 1.5"},
    {"display_name": "Depth Leres++", "name": "depth_leres++", "path": "lllyasviel/control_v11f1p_sd15_depth", "version": "SD 1.5"},
    {"display_name": "Depth Midas", "name": "depth_midas", "path": "lllyasviel/control_v11f1p_sd15_depth", "version": "SD 1.5"},
    {"display_name": "Depth Zoe", "name": "depth_zoe", "path": "lllyasviel/control_v11f1p_sd15_depth", "version": "SD 1.5"},
    {"display_name": "MLSD", "name": "mlsd", "path": "lllyasviel/control_v11p_sd15_mlsd", "version": "SD 1.5"},
    {"display_name": "Normal Bae", "name": "normal_bae", "path": "lllyasviel/control_v11p_sd15_normalbae", "version": "SD 1.5"},
    {"display_name": "Normal Midas", "name": "normal_midas", "path": "lllyasviel/control_v11p_sd15_normalbae", "version": "SD 1.5"},
    {"display_name": "Scribble Hed", "name": "scribble_hed", "path": "lllyasviel/control_v11p_sd15_scribble", "version": "SD 1.5"},
    {"display_name": "Scribble Pidinet", "name": "scribble_pidinet", "path": "lllyasviel/control_v11p_sd15_scribble", "version": "SD 1.5"},
    {"display_name": "Segmentation", "name": "segmentation", "path": "lllyasviel/control_v11p_sd15_seg", "version": "SD 1.5"},
    {"display_name": "Lineart Coarse", "name": "lineart_coarse", "path": "lllyasviel/control_v11p_sd15_lineart", "version": "SD 1.5"},
    {"display_name": "Lineart Realistic", "name": "lineart_realistic", "path": "lllyasviel/control_v11p_sd15_lineart", "version": "SD 1.5"},
    {"display_name": "Lineart Anime", "name": "lineart_anime", "path": "lllyasviel/control_v11p_sd15s2_lineart_anime", "version": "SD 1.5"},
    {"display_name": "Openpose", "name": "openpose", "path": "lllyasviel/control_v11p_sd15_openpose", "version": "SD 1.5"},
    {"display_name": "Openpose Face", "name": "openpose_face", "path": "lllyasviel/control_v11p_sd15_openpose", "version": "SD 1.5"},
    {"display_name": "Openpose Faceonly", "name": "openpose_faceonly", "path": "lllyasviel/control_v11p_sd15_openpose", "version": "SD 1.5"},
    {"display_name": "Openpose Full", "name": "openpose_full", "path": "lllyasviel/control_v11p_sd15_openpose", "version": "SD 1.5"},
    {"display_name": "Openpose Hand", "name": "openpose_hand", "path": "lllyasviel/control_v11p_sd15_openpose", "version": "SD 1.5"},
    {"display_name": "Scribble Hed", "name": "softedge_hed", "path": "lllyasviel/control_v11p_sd15_softedge", "version": "SD 1.5"},
    {"display_name": "Softedge Hedsafe", "name": "softedge_hedsafe", "path": "lllyasviel/control_v11p_sd15_softedge", "version": "SD 1.5"},
    {"display_name": "Scribble Pidinet", "name": "softedge_pidinet", "path": "lllyasviel/control_v11p_sd15_softedge", "version": "SD 1.5"},
    {"display_name": "Softedge Pidsafe", "name": "softedge_pidsafe", "path": "lllyasviel/control_v11p_sd15_softedge", "version": "SD 1.5"},
    {"display_name": "Pixel2Pixel", "name": "pixel2pixel", "path": "lllyasviel/control_v11e_sd15_ip2p", "version": "SD 1.5"},
    {"display_name": "Inpaint", "name": "inpaint", "path": "lllyasviel/control_v11p_sd15_inpaint", "version": "SD 1.5"},
    {"display_name": "Shuffle", "name": "shuffle", "path": "lllyasviel/control_v11e_sd15_shuffle", "version": "SD 1.5"},
    {"display_name": "Canny", "name": "canny", "path": "diffusers/controlnet-canny-sdxl-1.0-small", "version": "SDXL 1.0"},
]



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'chat_prompt.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QFrame, QGridLayout,
    QHBoxLayout, QPlainTextEdit, QProgressBar, QPushButton,
    QScrollArea, QSizePolicy, QSplitter, QTextEdit,
    QVBoxLayout, QWidget)

class Ui_chat_prompt(object):
    def setupUi(self, chat_prompt):
        if not chat_prompt.objectName():
            chat_prompt.setObjectName(u"chat_prompt")
        chat_prompt.resize(597, 691)
        self.gridLayout_3 = QGridLayout(chat_prompt)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.gridLayout_3.setHorizontalSpacing(0)
        self.gridLayout_3.setVerticalSpacing(10)
        self.gridLayout_3.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(5)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.action = QComboBox(chat_prompt)
        self.action.setObjectName(u"action")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.action.sizePolicy().hasHeightForWidth())
        self.action.setSizePolicy(sizePolicy)
        self.action.setMinimumSize(QSize(100, 30))
        self.action.setSizeAdjustPolicy(QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)

        self.horizontalLayout.addWidget(self.action)

        self.progressBar = QProgressBar(chat_prompt)
        self.progressBar.setObjectName(u"progressBar")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.progressBar.sizePolicy().hasHeightForWidth())
        self.progressBar.setSizePolicy(sizePolicy1)
        self.progressBar.setMinimumSize(QSize(0, 30))
        self.progressBar.setValue(0)

        self.horizontalLayout.addWidget(self.progressBar)

        self.send_button = QPushButton(chat_prompt)
        self.send_button.setObjectName(u"send_button")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.send_button.sizePolicy().hasHeightForWidth())
        self.send_button.setSizePolicy(sizePolicy2)
        self.send_button.setMinimumSize(QSize(30, 30))
        self.send_button.setMaximumSize(QSize(30, 30))
        icon = QIcon(QIcon.fromTheme(u"mail-reply-sender"))
        self.send_button.setIcon(icon)

        self.horizontalLayout.addWidget(self.send_button)

        self.clear_conversatiion_button = QPushButton(chat_prompt)
        self.clear_conversatiion_button.setObjectName(u"clear_conversatiion_button")
        sizePolicy2.setHeightForWidth(self.clear_conversatiion_button.sizePolicy().hasHeightForWidth())
        self.clear_conversatiion_button.setSizePolicy(sizePolicy2)
        self.clear_conversatiion_button.setMinimumSize(QSize(30, 30))
        self.clear_conversatiion_button.setMaximumSize(QSize(30, 30))
        icon1 = QIcon(QIcon.fromTheme(u"document-new"))
        self.clear_conversatiion_button.setIcon(icon1)

        self.horizontalLayout.addWidget(self.clear_conversatiion_button)

        self.pushButton = QPushButton(chat_prompt)
        self.pushButton.setObjectName(u"pushButton")
        sizePolicy2.setHeightForWidth(self.pushButton.sizePolicy().hasHeightForWidth())
        self.pushButton.setSizePolicy(sizePolicy2)
        self.pushButton.setMinimumSize(QSize(30, 30))
        self.pushButton.setMaximumSize(QSize(30, 30))
        icon2 = QIcon(QIcon.fromTheme(u"media-playback-stop"))
        self.pushButton.setIcon(icon2)

        self.horizontalLayout.addWidget(self.pushButton)


        self.gridLayout_3.addLayout(self.horizontalLayout, 2, 0, 1, 1)

        self.scrollArea = QScrollArea(chat_prompt)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self.scrollAreaWidgetContents_2 = QWidget()
        self.scrollAreaWidgetContents_2.setObjectName(u"scrollAreaWidgetContents_2")
        self.scrollAreaWidgetContents_2.setGeometry(QRect(0, 0, 595, 647))
        self.gridLayout = QGridLayout(self.scrollAreaWidgetContents_2)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setHorizontalSpacing(0)
        self.gridLayout.setVerticalSpacing(10)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.chat_prompt_splitter = QSplitter(self.scrollAreaWidgetContents_2)
        self.chat_prompt_splitter.setObjectName(u"chat_prompt_splitter")
        self.chat_prompt_splitter.setOrientation(Qt.Orientation.Vertical)
        self.chat_prompt_splitter.setChildrenCollapsible(False)
        self.verticalLayoutWidget = QWidget(self.chat_prompt_splitter)
        self.verticalLayoutWidget.setObjectName(u"verticalLayoutWidget")
        self.verticalLayout = QVBoxLayout(self.verticalLayoutWidget)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.chat_container = QScrollArea(self.verticalLayoutWidget)
        self.chat_container.setObjectName(u"chat_container")
        self.chat_container.setFrameShape(QFrame.Shape.NoFrame)
        self.chat_container.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 593, 321))
        self.gridLayout_2 = QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout_2.setSpacing(0)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.conversation = QTextEdit(self.scrollAreaWidgetContents)
        self.conversation.setObjectName(u"conversation")
        font = QFont()
        font.setFamilies([u"CCJoeKubert"])
        self.conversation.setFont(font)
        self.conversation.setReadOnly(True)

        self.gridLayout_2.addWidget(self.conversation, 0, 0, 1, 1)

        self.chat_container.setWidget(self.scrollAreaWidgetContents)

        self.verticalLayout.addWidget(self.chat_container)

        self.chat_prompt_splitter.addWidget(self.verticalLayoutWidget)
        self.prompt = QPlainTextEdit(self.chat_prompt_splitter)
        self.prompt.setObjectName(u"prompt")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.prompt.sizePolicy().hasHeightForWidth())
        self.prompt.setSizePolicy(sizePolicy3)
        self.prompt.setMinimumSize(QSize(0, 150))
        self.prompt.setMaximumSize(QSize(16777215, 16777215))
        self.chat_prompt_splitter.addWidget(self.prompt)

        self.gridLayout.addWidget(self.chat_prompt_splitter, 0, 0, 1, 1)

        self.scrollArea.setWidget(self.scrollAreaWidgetContents_2)

        self.gridLayout_3.addWidget(self.scrollArea, 0, 0, 1, 1)


        self.retranslateUi(chat_prompt)
        self.prompt.textChanged.connect(chat_prompt.prompt_text_changed)
        self.action.currentTextChanged.connect(chat_prompt.llm_action_changed)
        self.send_button.clicked.connect(chat_prompt.action_button_clicked_send)
        self.clear_conversatiion_button.clicked.connect(chat_prompt.action_button_clicked_clear_conversation)
        self.pushButton.clicked.connect(chat_prompt.interrupt_button_clicked)

        QMetaObject.connectSlotsByName(chat_prompt)
    # setupUi

    def retranslateUi(self, chat_prompt):
        chat_prompt.setWindowTitle(QCoreApplication.translate("chat_prompt", u"Form", None))
#if QT_CONFIG(tooltip)
        self.send_button.setToolTip(QCoreApplication.translate("chat_prompt", u"Send message", None))
#endif // QT_CONFIG(tooltip)
        self.send_button.setText("")
#if QT_CONFIG(tooltip)
        self.clear_conversatiion_button.setToolTip(QCoreApplication.translate("chat_prompt", u"Clear conversation", None))
#endif // QT_CONFIG(tooltip)
        self.clear_conversatiion_button.setText("")
#if QT_CONFIG(tooltip)
        self.pushButton.setToolTip(QCoreApplication.translate("chat_prompt", u"Cancel message", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton.setText("")
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'llama_license.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QGridLayout, QLabel,
    QScrollArea, QSizePolicy, QWidget)

class Ui_llama_license(object):
    def setupUi(self, llama_license):
        if not llama_license.objectName():
            llama_license.setObjectName(u"llama_license")
        llama_license.resize(568, 603)
        self.gridLayout_2 = QGridLayout(llama_license)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.scrollArea = QScrollArea(llama_license)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 739, 2007))
        self.gridLayout = QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout.setObjectName(u"gridLayout")
        self.label_2 = QLabel(self.scrollAreaWidgetContents)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignVCenter)

        self.gridLayout.addWidget(self.label_2, 0, 0, 1, 1)

        self.scrollArea.setWidget(self.scrollAreaWidgetContents)

        self.gridLayout_2.addWidget(self.scrollArea, 1, 0, 1, 1)

        self.label = QLabel(llama_license)
        self.label.setObjectName(u"label")
        font = QFont()
        font.setPointSize(11)
        font.setBold(True)
        self.label.setFont(font)

        self.gridLayout_2.addWidget(self.label, 0, 0, 1, 1)

        self.checkBox = QCheckBox(llama_license)
        self.checkBox.setObjectName(u"checkBox")

        self.gridLayout_2.addWidget(self.checkBox, 2, 0, 1, 1)


        self.retranslateUi(llama_license)
        self.checkBox.clicked["bool"].connect(llama_license.agreement_clicked)

        QMetaObject.connectSlotsByName(llama_license)
    # setupUi

    def retranslateUi(self, llama_license):
        llama_license.setWindowTitle(QCoreApplication.translate("llama_license", u"Form", None))
        self.label_2.setText(QCoreApplication.translate("llama_license", u"META LLAMA 3 COMMUNITY LICENSE AGREEMENT\n"
"Meta Llama 3 Version Release Date: April 18, 2024\n"
"\n"
"\u201cAgreement\u201d means the terms and conditions for use, reproduction, distribution and modification of the\n"
"Llama Materials set forth herein.\n"
"\n"
"\u201cDocumentation\u201d means the specifications, manuals and documentation accompanying Meta Llama 3\n"
"distributed by Meta at https://llama.meta.com/get-started/.\n"
"\n"
"\u201cLicensee\u201d or \u201cyou\u201d means you, or your employer or any other person or entity (if you are entering into\n"
"this Agreement on such person or entity\u2019s behalf), of the age required under applicable laws, rules or\n"
"regulations to provide legal consent and that has legal authority to bind your employer or such other\n"
"person or entity if you are entering in this Agreement on their behalf.\n"
"\n"
"\u201cMeta Llama 3\u201d means the foundational large language models and software and algorithms, including\n"
"machine-learning model code, trained model w"
                        "eights, inference-enabling code, training-enabling code,\n"
"fine-tuning enabling code and other elements of the foregoing distributed by Meta at\n"
"https://llama.meta.com/llama-downloads.\n"
"\n"
"\u201cLlama Materials\u201d means, collectively, Meta\u2019s proprietary Meta Llama 3 and Documentation (and any\n"
"portion thereof) made available under this Agreement.\n"
"\n"
"\u201cMeta\u201d or \u201cwe\u201d means Meta Platforms Ireland Limited (if you are located in or, if you are an entity, your\n"
"principal place of business is in the EEA or Switzerland) and Meta Platforms, Inc. (if you are located\n"
"outside of the EEA or Switzerland).\n"
"\n"
"By clicking \u201cI Accept\u201d below or by using or distributing any portion or element of the Llama Materials,\n"
"you agree to be bound by this Agreement.\n"
"\n"
"1. License Rights and Redistribution.\n"
"\n"
"  a. Grant of Rights. You are granted a non-exclusive, worldwide, non-transferable and royalty-free\n"
"limited license under Meta\u2019s intellectua"
                        "l property or other rights owned by Meta embodied in the Llama\n"
"Materials to use, reproduce, distribute, copy, create derivative works of, and make modifications to the\n"
"Llama Materials.\n"
"\n"
"  b. Redistribution and Use.\n"
"\n"
"      i. If you distribute or make available the Llama Materials (or any derivative works\n"
"thereof), or a product or service that uses any of them, including another AI model, you shall (A) provide\n"
"a copy of this Agreement with any such Llama Materials; and (B) prominently display \u201cBuilt with Meta\n"
"Llama 3\u201d on a related website, user interface, blogpost, about page, or product documentation. If you\n"
"use the Llama Materials to create, train, fine tune, or otherwise improve an AI model, which is\n"
"distributed or made available, you shall also include \u201cLlama 3\u201d at the beginning of any such AI model\n"
"name.\n"
"\n"
"      ii. If you receive Llama Materials, or any derivative works thereof, from a Licensee as part \n"
"of an integrated end use"
                        "r product, then Section 2 of this Agreement will not apply to you.\n"
"\n"
"      iii. You must retain in all copies of the Llama Materials that you distribute the following\n"
"attribution notice within a \u201cNotice\u201d text file distributed as a part of such copies: \u201cMeta Llama 3 is\n"
"licensed under the Meta Llama 3 Community License, Copyright \u00a9 Meta Platforms, Inc. All Rights\n"
"Reserved.\u201d\n"
"\n"
"      iv. Your use of the Llama Materials must comply with applicable laws and regulations\n"
"(including trade compliance laws and regulations) and adhere to the Acceptable Use Policy for the Llama\n"
"Materials (available at https://llama.meta.com/llama3/use-policy), which is hereby incorporated by\n"
"reference into this Agreement.\n"
"\n"
"      v. You will not use the Llama Materials or any output or results of the Llama Materials to\n"
"improve any other large language model (excluding Meta Llama 3 or derivative works thereof).\n"
"\n"
"2. Additional Commercial Terms. If, on the Meta "
                        "Llama 3 version release date, the monthly active users\n"
"of the products or services made available by or for Licensee, or Licensee\u2019s affiliates, is greater than 700\n"
"million monthly active users in the preceding calendar month, you must request a license from Meta,\n"
"which Meta may grant to you in its sole discretion, and you are not authorized to exercise any of the\n"
"rights under this Agreement unless or until Meta otherwise expressly grants you such rights.\n"
"\n"
"3. Disclaimer of Warranty. UNLESS REQUIRED BY APPLICABLE LAW, THE LLAMA MATERIALS AND ANY\n"
"OUTPUT AND RESULTS THEREFROM ARE PROVIDED ON AN \u201cAS IS\u201d BASIS, WITHOUT WARRANTIES OF\n"
"ANY KIND, AND META DISCLAIMS ALL WARRANTIES OF ANY KIND, BOTH EXPRESS AND IMPLIED,\n"
"INCLUDING, WITHOUT LIMITATION, ANY WARRANTIES OF TITLE, NON-INFRINGEMENT,\n"
"MERCHANTABILITY, OR FITNESS FOR A PARTICULAR PURPOSE. YOU ARE SOLELY RESPONSIBLE FOR\n"
"DETERMINING THE APPROPRIATENESS OF USING OR REDISTRIBUTING THE LLAMA MATERIALS AND\n"
"AS"
                        "SUME ANY RISKS ASSOCIATED WITH YOUR USE OF THE LLAMA MATERIALS AND ANY OUTPUT AND\n"
"RESULTS.\n"
"\n"
"4. Limitation of Liability. IN NO EVENT WILL META OR ITS AFFILIATES BE LIABLE UNDER ANY THEORY OF\n"
"LIABILITY, WHETHER IN CONTRACT, TORT, NEGLIGENCE, PRODUCTS LIABILITY, OR OTHERWISE, ARISING\n"
"OUT OF THIS AGREEMENT, FOR ANY LOST PROFITS OR ANY INDIRECT, SPECIAL, CONSEQUENTIAL,\n"
"INCIDENTAL, EXEMPLARY OR PUNITIVE DAMAGES, EVEN IF META OR ITS AFFILIATES HAVE BEEN ADVISED\n"
"OF THE POSSIBILITY OF ANY OF THE FOREGOING.\n"
"\n"
"5. Intellectual Property.\n"
"\n"
"  a. No trademark licenses are granted under this Agreement, and in connection with the Llama\n"
"Materials, neither Meta nor Licensee may use any name or mark owned by or associated with the other\n"
"or any of its affiliates, except as required for reasonable and customary use in describing and\n"
"redistributing the Llama Materials or as set forth in this Section 5(a). Meta hereby grants you a license to\n"
"use \u201cLlama 3\u201d (the \u201c"
                        "Mark\u201d) solely as required to comply with the last sentence of Section 1.b.i. You will\n"
"comply with Meta\u2019s brand guidelines (currently accessible at\n"
"https://about.meta.com/brand/resources/meta/company-brand/ ). All goodwill arising out of your use\n"
"of the Mark will inure to the benefit of Meta.\n"
"\n"
"  b. Subject to Meta\u2019s ownership of Llama Materials and derivatives made by or for Meta, with\n"
"respect to any derivative works and modifications of the Llama Materials that are made by you, as\n"
"between you and Meta, you are and will be the owner of such derivative works and modifications.\n"
"\n"
"  c. If you institute litigation or other proceedings against Meta or any entity (including a\n"
"cross-claim or counterclaim in a lawsuit) alleging that the Llama Materials or Meta Llama 3 outputs or\n"
"results, or any portion of any of the foregoing, constitutes infringement of intellectual property or other\n"
"rights owned or licensable by you, then any licenses granted to you under "
                        "this Agreement shall\n"
"terminate as of the date such litigation or claim is filed or instituted. You will indemnify and hold\n"
"harmless Meta from and against any claim by any third party arising out of or related to your use or\n"
"distribution of the Llama Materials.\n"
"\n"
"6. Term and Termination. The term of this Agreement will commence upon your acceptance of this\n"
"Agreement or access to the Llama Materials and will continue in full force and effect until terminated in\n"
"accordance with the terms and conditions herein. Meta may terminate this Agreement if you are in\n"
"breach of any term or condition of this Agreement. Upon termination of this Agreement, you shall delete\n"
"and cease use of the Llama Materials. Sections 3, 4 and 7 shall survive the termination of this\n"
"Agreement.\n"
"\n"
"7. Governing Law and Jurisdiction. This Agreement will be governed and construed under the laws of\n"
"the State of California without regard to choice of law principles, and the UN Convention on Contracts"
                        "\n"
"for the International Sale of Goods does not apply to this Agreement. The courts of California shall have\n"
"exclusive jurisdiction of any dispute arising out of this Agreement.", None))
        self.label.setText(QCoreApplication.translate("llama_license", u"META LLAMA 3 COMMUNITY LICENSE AGREEMENT", None))
        self.checkBox.setText(QCoreApplication.translate("llama_license", u"I have read and agree to the terms of the Meta LLama 3 License", None))
    # retranslateUi




import numpy as np
from PIL import Image
from PySide6.QtCore import QRect
from airunner.enums import SignalCode
from airunner.utils.image.convert_binary_to_image import convert_binary_to_image
from airunner.workers.worker import Worker


class MaskGeneratorWorker(Worker):
    @property
    def active_rect(self):
        rect = QRect(
            self.active_grid_settings.pos_x,
            self.active_grid_settings.pos_y,
            self.application_settings.working_width,
            self.application_settings.working_height
        )
        rect.translate(-self.drawing_pad_settings.x_pos, -self.drawing_pad_settings.y_pos)
        return rect

    def register_signals(self):
        self.register(SignalCode.GENERATE_MASK, self.on_generate_mask_signal)

    def on_generate_mask_signal(self, message: dict):
        mask = self.generate_mask()
        self.emit_signal(
            SignalCode.MASK_GENERATOR_WORKER_RESPONSE_SIGNAL,
            {"mask": mask}
        )

    def generate_mask(self) -> Image:
        base_64_image = self.drawing_pad_settings.image
        if base_64_image is None:
            return
        image = convert_binary_to_image(base_64_image)
        image = image.convert('RGBA')
        if image is not None:
            image_width = image.width
            image_height = image.height

            # Create a white image of the same size as the working area
            mask = Image.new('RGB', (self.application_settings.working_width, self.application_settings.working_height),
                             'white')
            mask_data = np.array(mask)

            # Calculate the position and size of the black rectangle (image)
            black_left = max(0, -self.active_rect.left())
            black_top = max(0, -self.active_rect.top())
            black_right = min(image_width, self.application_settings.working_width - self.active_rect.left())
            black_bottom = min(image_height, self.application_settings.working_height - self.active_rect.top())

            # Ensure the coordinates are valid
            if black_right >= black_left and black_bottom >= black_top:
                # Draw the black rectangle (image) on the mask
                mask_data[black_top:black_bottom, black_left:black_right] = [0, 0, 0]

            # Check for transparent pixels and draw them as white on the mask
            if image.mode == 'RGBA':
                image_data = np.array(image)
                transparent_pixels = (image_data[:, :, 3] == 0)

                # Calculate the mask coordinates for transparent pixels
                mask_y, mask_x = np.where(transparent_pixels)  # Note the order of mask_x and mask_y
                mask_x -= self.active_rect.left()
                mask_y -= self.active_rect.top()

                # Ensure the coordinates are within the mask bounds
                valid_coords = (mask_x >= 0) & (mask_x < self.application_settings.working_width) & \
                               (mask_y >= 0) & (mask_y < self.application_settings.working_height)
                mask_x = mask_x[valid_coords]
                mask_y = mask_y[valid_coords]

                # Draw white pixels on the mask
                mask_data[mask_y, mask_x] = [255, 255, 255]

            # Convert the modified mask data back to an image
            mask = Image.fromarray(mask_data)
            return mask
        return None



import pyttsx3
from airunner.handlers.tts.tts_handler import TTSHandler
from airunner.enums import ModelType, ModelStatus, Gender


class EspeakTTSHandler(TTSHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._rate = None
        self._pitch = None
        self._volume = None
        self._voice = None
        self._language = None
        self._gender = None

    def generate(self, message:str):
        message = message.replace('"', "'")
        if message != "":
            self._engine.say(message)
            self._engine.runAndWait()
        return None

    def load(self, target_model=None):
        self.logger.debug("Initializing espeak")
        self.unload()
        self.change_model_status(ModelType.TTS, ModelStatus.LOADING)
        self._engine = pyttsx3.init()
        self._initialize()
        self.change_model_status(ModelType.TTS, ModelStatus.LOADED)

    def unload(self):
        self.logger.debug("Unloading espeak")
        self.change_model_status(ModelType.TTS, ModelStatus.LOADING)
        self._engine = None
        self.change_model_status(ModelType.TTS, ModelStatus.UNLOADED)

    def unblock_tts_generator_signal(self):
        pass

    def interrupt_process_signal(self):
        pass

    def _initialize(self):
        self._rate = self.espeak_settings.rate
        self._pitch = self.espeak_settings.pitch
        self._volume = self.espeak_settings.volume
        self._voice = self.espeak_settings.voice
        self._language = self.espeak_settings.language
        gender = self.espeak_settings.gender
        if gender != self.espeak_settings.gender:
            self._gender = gender
            self._engine.setProperty(
                'gender',
                Gender(self._gender)
            )
        self._engine.setProperty('rate', float(self._rate))
        self._engine.setProperty('volume', self._volume / 100.0)
        self._engine.setProperty('pitch', float(self._pitch))
        self._engine.setProperty('voice', f'{self._voice}')
        self._engine.setProperty('language', self._language)



import numpy as np
from PIL import Image
from airunner.enums import (
    SignalCode,
    SDMode,
)
from airunner.workers.worker import Worker
SKIP_RELOAD_CONSTS = (
    SDMode.FAST_GENERATE,
    SDMode.DRAWING,
)


class LatentsWorker(Worker):
    def __init__(self):
        super().__init__()
        self.register(SignalCode.HANDLE_LATENTS_SIGNAL, self.on_handle_latents_signal)

    def on_handle_latents_signal(self, data: dict):
        latents = data.get("latents")
        sd_request = data.get("sd_request")
        # convert latents to PIL image
        latents = latents[0].detach().cpu().numpy().astype(np.uint8)  # convert to uint8
        latents = latents.transpose(1, 2, 0)
        image = Image.fromarray(latents)
        image = image.resize((self.application_settings.working_width, self.application_settings.working_height))
        image = image.convert("RGBA")
        self.emit_signal(
            SignalCode.SD_IMAGE_GENERATED_SIGNAL,
            {
                "images": [image],
                "action": sd_request.section,
                "outpaint_box_rect": sd_request.active_rect,
            }
        )
        self.sd_request = None



import os

from airunner.data.models.settings_models import Lora, Embedding
from airunner.windows.main.settings_mixin import SettingsMixin


def scan_path_for_lora(base_path) -> bool:
    lora_added = False
    lora_deleted = False

    db_handler = SettingsMixin()
    for versionpath, versionnames, versionfiles in os.walk(os.path.expanduser(os.path.join(base_path, "art/models"))):
        version = versionpath.split("/")[-1]
        lora_path = os.path.expanduser(
            os.path.join(
                base_path,
                "art/models",
                version,
                "lora"
            )
        )
        if not os.path.exists(lora_path):
            continue

        existing_lora = db_handler.session.query(Lora).all()
        for lora in existing_lora:
            if not os.path.exists(lora.path):
                db_handler.session.delete(lora)
                lora_deleted = True
        for dirpath, dirnames, filenames in os.walk(lora_path):
            for file in filenames:
                if file.endswith(".ckpt") or file.endswith(".safetensors") or file.endswith(".pt"):
                    name = file.replace(".ckpt", "").replace(".safetensors", "").replace(".pt", "")
                    path = os.path.join(dirpath, file)
                    item = db_handler.get_lora_by_name(name)
                    if not item or item.path != path or item.version != version:
                        item = Lora(
                            name=name,
                            path=path,
                            scale=1,
                            enabled=False,
                            loaded=False,
                            trigger_word="",
                            version=version
                        )
                        db_handler.session.add(item)
                        lora_added = True
        if lora_deleted or lora_added:
            db_handler.session.commit()
    return lora_deleted or lora_added

def scan_path_for_embeddings(base_path) -> bool:
    embedding_added = False
    embedding_deleted = False
    db_handler = SettingsMixin()
    items = []
    for versionpath, versionnames, versionfiles in os.walk(os.path.expanduser(os.path.join(base_path, "art/models"))):
        version = versionpath.split("/")[-1]
        embedding_path = os.path.expanduser(
            os.path.join(
                base_path,
                "art/models",
                version,
                "embeddings"
            )
        )
        if not os.path.exists(embedding_path):
            continue
        existing_embeddings = db_handler.session.query(Embedding).all()
        for embedding in existing_embeddings:
            if not os.path.exists(embedding.path):
                db_handler.session.delete(embedding)
                embedding_deleted = True
        for dirpath, dirnames, filenames in os.walk(embedding_path):
            for file in filenames:
                if file.endswith(".ckpt") or file.endswith(".safetensors") or file.endswith(".pt"):
                    name = file.replace(".ckpt", "").replace(".safetensors", "").replace(".pt", "")
                    path = os.path.join(dirpath, file)
                    item = db_handler.get_embedding_by_name(name)
                    if not item or item.path != path or item.version != version:
                        item = Embedding(
                            name=name,
                            path=path,
                            version=version,
                            tags="",
                            active=False,
                            trigger_word=""
                        )
                        db_handler.session.add(item)
                        embedding_added = True
        if embedding_deleted or embedding_added:
            db_handler.session.commit()
    return embedding_deleted or embedding_added



"""Add lock_input_image column to image_to_image_settings

Revision ID: 6d58cbfd61fb
Revises: 72d9134823cb
Create Date: 2024-10-09 10:05:11.349862

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

from airunner.utils.db.column_exists import column_exists

# revision identifiers, used by Alembic.
revision: str = '6d58cbfd61fb'
down_revision: Union[str, None] = '72d9134823cb'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade():
    if not column_exists('image_to_image_settings', 'lock_input_image'):
        op.add_column('image_to_image_settings', sa.Column('lock_input_image', sa.Boolean, default=False))
    if not column_exists('controlnet_settings', 'lock_input_image'):
        op.add_column('controlnet_settings', sa.Column('lock_input_image', sa.Boolean, default=False))

def downgrade():
    if column_exists('image_to_image_settings', 'lock_input_image'):
        op.drop_column('image_to_image_settings', 'lock_input_image')
    if column_exists('controlnet_settings', 'lock_input_image'):
        op.drop_column('image_to_image_settings', 'lock_input_image')
    # ### end Alembic commands ###



#!/usr/bin/env python

# import facehuggershield
# facehuggershield.huggingface.activate(
#     show_stdout=True,
#     darklock_os_whitelisted_operations=[
#         "makedirs"
#     ]
# )
from airunner.app_installer import AppInstaller


if __name__ == "__main__":
    AppInstaller()



"""Create whisper_settings table

Revision ID: 536e18463461
Revises: 3212e8fccc68
Create Date: 2024-10-13 09:34:20.974389

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

from airunner.data.models.settings_models import WhisperSettings

# revision identifiers, used by Alembic.
revision: str = '536e18463461'
down_revision: Union[str, None] = '3212e8fccc68'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None

def set_default_values(model_name_):
    default_values = {}
    for column in model_name_.__table__.columns:
        if column.default is not None:
            default_values[column.name] = column.default.arg
    op.bulk_insert(
        model_name_.__table__,
        [default_values]
    )

def upgrade():
    try:
        op.create_table(
            WhisperSettings.__tablename__,
            *WhisperSettings.__table__.columns
        )
    except Exception as e:
        print(f"Error: {e}")

    try:
        set_default_values(WhisperSettings)
    except Exception as e:
        print(f"Error: {e}")

def downgrade():
    try:
        op.drop_table('whisper_settings')
    except Exception as e:
        print(f"Error: {e}")
    # ### end Alembic commands ###



"""Modify columns in stt_settings table

Revision ID: 3212e8fccc68
Revises: 6d58cbfd61fb
Create Date: 2024-10-13 09:29:29.878635

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision: str = '3212e8fccc68'
down_revision: Union[str, None] = '6d58cbfd61fb'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None

def upgrade():
    # Modify columns to Float if they are not already Float
    try:
        with op.batch_alter_table('stt_settings') as batch_op:
            batch_op.alter_column('volume_input_threshold', type_=sa.Float, existing_type=sa.Integer)
            batch_op.alter_column('silence_buffer_seconds', type_=sa.Float, existing_type=sa.Integer)
            batch_op.alter_column('chunk_duration', type_=sa.Float, existing_type=sa.Integer)
    except sqlite.DatabaseError:
        pass

def downgrade():
    # Revert columns back to Integer
    try:
        with op.batch_alter_table('stt_settings') as batch_op:
            batch_op.alter_column('volume_input_threshold', type_=sa.Integer, existing_type=sa.Float)
            batch_op.alter_column('silence_buffer_seconds', type_=sa.Integer, existing_type=sa.Float)
            batch_op.alter_column('chunk_duration', type_=sa.Integer, existing_type=sa.Float)
    except sqlite.DatabaseError:
        pass
    # ### end Alembic commands ###



import os
import re
import sys
import urllib
import webbrowser
from functools import partial

import requests
from PIL import Image
from PySide6 import QtGui
from PySide6.QtCore import (
    Slot,
    Signal,
    QProcess
)
from PySide6.QtGui import QGuiApplication, QKeySequence
from PySide6.QtWidgets import (
    QApplication,
    QMainWindow,
    QMessageBox,
    QCheckBox, QInputDialog
)
from bs4 import BeautifulSoup

from airunner.handlers.llm.agent.actions.bash_execute import bash_execute
from airunner.handlers.llm.agent.actions.show_path import show_path
from airunner.data.models.settings_models import ShortcutKeys, ImageFilter, DrawingPadSettings
from airunner.app_installer import AppInstaller
from airunner.settings import (
    STATUS_ERROR_COLOR,
    STATUS_NORMAL_COLOR_LIGHT,
    STATUS_NORMAL_COLOR_DARK,
    NSFW_CONTENT_DETECTED_MESSAGE,
)
from airunner.enums import (
    SignalCode,
    CanvasToolName,
    GeneratorSection,
    LLMAction, ModelType, ModelStatus
)
from airunner.mediator_mixin import MediatorMixin
from airunner.settings import (
    BASE_PATH,
    BUG_REPORT_LINK,
    VULNERABILITY_REPORT_LINK
)
from airunner.styles_mixin import StylesMixin
from airunner.utils.image.convert_image_to_binary import convert_image_to_binary
from airunner.utils.create_worker import create_worker
from airunner.workers.audio_capture_worker import AudioCaptureWorker
from airunner.workers.audio_processor_worker import AudioProcessorWorker
from airunner.workers.llm_generate_worker import LLMGenerateWorker
from airunner.workers.mask_generator_worker import MaskGeneratorWorker
from airunner.workers.sd_worker import SDWorker
from airunner.workers.tts_generator_worker import TTSGeneratorWorker
from airunner.workers.tts_vocalizer_worker import TTSVocalizerWorker

from airunner.utils.get_version import get_version
from airunner.utils.set_widget_state import set_widget_state
from airunner.widgets.model_manager.model_manager_widget import ModelManagerWidget
from airunner.widgets.stats.stats_widget import StatsWidget
from airunner.widgets.status.status_widget import StatusWidget
from airunner.windows.about.about import AboutWindow
from airunner.windows.filter_window import FilterWindow
from airunner.windows.main.ai_model_mixin import AIModelMixin
from airunner.windows.main.pipeline_mixin import PipelineMixin
from airunner.windows.main.settings_mixin import SettingsMixin
from airunner.windows.main.templates.main_window_ui import Ui_MainWindow
from airunner.windows.prompt_browser.prompt_browser import PromptBrowser
from airunner.windows.settings.airunner_settings import SettingsWindow
from airunner.windows.update.update_window import UpdateWindow


class MainWindow(
    QMainWindow,
    MediatorMixin,
    SettingsMixin,
    StylesMixin,
    PipelineMixin,
    AIModelMixin
):
    show_grid_toggled = Signal(bool)
    cell_size_changed_signal = Signal(int)
    line_width_changed_signal = Signal(int)
    line_color_changed_signal = Signal(str)
    canvas_color_changed_signal = Signal(str)
    snap_to_grid_changed_signal = Signal(bool)
    image_generated = Signal(bool)
    generator_tab_changed_signal = Signal()
    tab_section_changed_signal = Signal()
    load_image = Signal(str)
    load_image_object = Signal(object)
    loaded = Signal()
    window_opened = Signal()
    ui_class_ = Ui_MainWindow
    _window_title = f"AI Runner"
    icons = [
        ("pencil-icon", "actionToggle_Brush"),
        ("eraser-icon", "actionToggle_Eraser"),
        ("frame-grid-icon", "actionToggle_Grid"),
        ("circle-center-icon", "actionRecenter"),
        ("setting-line-icon", "actionSettings"),
        ("object-selected-icon", "actionToggle_Active_Grid_Area"),
        ("select-svgrepo-com", "actionToggle_Selection"),
        ("layer-icon", "actionMask_toggle"),
    ]

    def __init__(
        self,
        *args,
        defendatron=None,
        **kwargs
    ):
        self.ui = self.ui_class_()
        self.defendatron = defendatron
        self.quitting = False
        self.update_popup = None
        self._document_path = None
        self.prompt = None
        self.negative_prompt = None
        self.image_path = None
        self.token_signal = Signal(str)
        self.api = None
        self.input_event_manager = None
        self.current_filter = None
        self.tqdm_callback_triggered = False
        self.is_saved = False
        self.action = GeneratorSection.TXT2IMG.value
        self.progress_bar_started = False
        self.window = None
        self.canvas = None
        self.models = None
        self.client = None
        self._version = None
        self._latest_version = None
        self.data = None  # this is set in the generator_mixin image_handler function and used for deterministic generation
        self.status_error_color = STATUS_ERROR_COLOR
        self.status_normal_color_light = STATUS_NORMAL_COLOR_LIGHT
        self.status_normal_color_dark = STATUS_NORMAL_COLOR_DARK
        self._themes = None
        self.button_clicked_signal = Signal(dict)
        self.status_widget = None
        self.header_widget_spacer = None
        self.deterministic_window = None
        self.generator = None
        self._generator = None
        self._generator_settings = None
        self.listening = False
        self.initialized = False
        self._model_status = {model_type: ModelStatus.UNLOADED for model_type in ModelType}
        MediatorMixin.__init__(self)
        
        self.logger.debug("Starting AI Runnner")
        super().__init__(*args, **kwargs)
        self._updating_settings = True
        PipelineMixin.__init__(self)
        AIModelMixin.__init__(self)
        self._updating_settings = False
        self._worker_manager = None
        self.register_signals()
        self.initialize_ui()
        self._initialize_workers()

    @property
    def generator_tab_widget(self):
        return self.ui.generator_widget

    @property
    def version(self):
        if self._version is None:
            self._version = get_version()
        return f"v{self._version}"

    @property
    def latest_version(self):
        return self._latest_version

    @latest_version.setter
    def latest_version(self, val):
        self._latest_version = val

    @property
    def document_name(self):
        return "Untitled"

    @property
    def current_tool(self):
        return CanvasToolName(self.application_settings.current_tool)

    """
    Slot functions
    
    The following functions are defined in and connected to the appropriate
    signals in the corresponding ui file.
    """
    @Slot(bool)
    def action_toggle_brush(self, active: bool):
        self.toggle_tool(CanvasToolName.BRUSH, active)

    @Slot(bool)
    def action_toggle_eraser(self, active: bool):
        self.toggle_tool(CanvasToolName.ERASER, active)

    @Slot(bool)
    def action_toggle_select(self, active: bool):
        self.toggle_tool(CanvasToolName.SELECTION, active)

    @Slot(bool)
    def action_toggle_active_grid_area(self, active: bool):
        self.toggle_tool(CanvasToolName.ACTIVE_GRID_AREA, active)

    @Slot(bool)
    def action_toggle_nsfw_filter_triggered(self, val: bool):
        if val is False:
            self.show_nsfw_warning_popup()
        else:
            self.update_application_settings("nsfw_filter", val)
            self.toggle_nsfw_filter()
            self.emit_signal(SignalCode.SAFETY_CHECKER_LOAD_SIGNAL)

    @Slot()
    def action_center_clicked(self):
        print("center clicked")

    @Slot()
    def action_reset_settings(self):
        reply = QMessageBox.question(
            self,
            'Reset Settings',
            'Are you sure you want to reset all settings to their default values?',
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No
        )

        if reply == QMessageBox.Yes:
            self.reset_settings()
            self.restart()

    @Slot()
    def import_controlnet_image(self):
        self.emit_signal(SignalCode.CONTROLNET_IMPORT_IMAGE_SIGNAL)

    @Slot()
    def export_controlnet_image(self):
        self.emit_signal(SignalCode.CONTROLNET_EXPORT_IMAGE_SIGNAL)

    @Slot()
    def import_drawingpad_image(self):
        self.emit_signal(SignalCode.DRAWINGPAD_IMPORT_IMAGE_SIGNAL)

    @Slot()
    def export_drawingpad_image(self):
        self.emit_signal(SignalCode.DRAWINGPAD_EXPORT_IMAGE_SIGNAL)

    @Slot()
    def action_export_image_triggered(self):
        self.emit_signal(SignalCode.CANVAS_EXPORT_IMAGE_SIGNAL)

    @Slot()
    def action_import_image_triggered(self):
        self.emit_signal(SignalCode.CANVAS_IMPORT_IMAGE_SIGNAL)

    @Slot()
    def action_new_document_triggered(self):
        self.new_document()
        self.emit_signal(SignalCode.CANVAS_CLEAR)

    @Slot()
    def action_undo_triggered(self):
        self.emit_signal(SignalCode.UNDO_SIGNAL)

    @Slot()
    def action_redo_triggered(self):
        self.emit_signal(SignalCode.REDO_SIGNAL)

    @Slot()
    def action_paste_image_triggered(self):
        self.emit_signal(SignalCode.CANVAS_PASTE_IMAGE_SIGNAL)

    @Slot()
    def action_copy_image_triggered(self):
        self.emit_signal(SignalCode.CANVAS_COPY_IMAGE_SIGNAL)

    @Slot()
    def action_cut_image_triggered(self):
        self.emit_signal(SignalCode.CANVAS_CUT_IMAGE_SIGNAL)

    @Slot()
    def action_rotate_90_clockwise_triggered(self):
        self.emit_signal(SignalCode.CANVAS_ROTATE_90_CLOCKWISE_SIGNAL)

    @Slot()
    def action_rotate_90_counterclockwise_triggered(self):
        self.emit_signal(SignalCode.CANVAS_ROTATE_90_COUNTER_CLOCKWISE_SIGNAL)

    @Slot()
    def action_show_prompt_browser_triggered(self):
        PromptBrowser()

    @Slot()
    def action_clear_all_prompts_triggered(self):
        self.clear_all_prompts()

    @Slot()
    def action_show_model_manager(self):
        ModelManagerWidget()

    @Slot()
    def action_triggered_browse_ai_runner_path(self):
        path = self.path_settings.base_path
        if path == "":
            path = BASE_PATH
        show_path(path)

    @Slot()
    def action_show_images_path(self):
        self.show_settings_path("image_path")

    @Slot()
    def action_show_model_path_txt2img(self):
        self.show_settings_path("txt2img_model_path")

    @Slot()
    def action_show_model_path_inpaint(self):
        self.show_settings_path("inpaint_model_path")

    @Slot()
    def action_show_model_path_embeddings(self):
        self.show_settings_path("embeddings_model_path")

    @Slot()
    def action_show_model_path_lora(self):
        self.show_settings_path("lora_model_path")

    @Slot()
    def action_show_llm(self):
        pass

    @Slot()
    def action_show_about_window(self):
        AboutWindow()

    @Slot()
    def action_show_settings(self):
        SettingsWindow()

    @Slot()
    def action_open_vulnerability_report(self):
        webbrowser.open(VULNERABILITY_REPORT_LINK)

    @Slot()
    def action_open_bug_report(self):
        webbrowser.open(BUG_REPORT_LINK)

    @Slot()
    def action_open_discord(self):
        pass

    @Slot(bool)
    def action_toggle_mask_layer(self, val: bool):
        if val is True and self.drawing_pad_mask is None:
            self._generate_drawingpad_mask()

        self.update_drawing_pad_settings("mask_layer_enabled", val)
        self.emit_signal(SignalCode.MASK_LAYER_TOGGLED)

    @Slot(bool)
    def action_outpaint_toggled(self, val: bool):
        self.update_outpaint_settings("enabled", val)

    @Slot()
    def action_outpaint_export(self):
        self.emit_signal(SignalCode.OUTPAINT_EXPORT_SIGNAL)

    @Slot()
    def action_outpaint_import(self):
        self.emit_signal(SignalCode.OUTPAINT_IMPORT_SIGNAL)

    @Slot()
    def action_run_setup_wizard_clicked(self):
        self.show_setup_wizard()

    @Slot(bool)
    def action_toggle_llm(self, val: bool):
        self.on_toggle_llm(val=val)

    @Slot(bool)
    def action_image_generator_toggled(self, val: bool):
        self.on_toggle_sd(val=val)

    @Slot(bool)
    def tts_button_toggled(self, val: bool):
        self.on_toggle_tts(val=val)

    @Slot(bool)
    def action_controlnet_toggled(self, val: bool):
        self._update_action_button(
            ModelType.CONTROLNET,
            self.ui.actionToggle_Controlnet,
            val,
            SignalCode.CONTROLNET_LOAD_SIGNAL,
            SignalCode.CONTROLNET_UNLOAD_SIGNAL,
            "controlnet_enabled"
        )

    @Slot(bool)
    def v2t_button_toggled(self, val):
        if self._model_status[ModelType.STT] is ModelStatus.LOADING:
            val = not val
        self._update_action_button(
            ModelType.STT,
            self.ui.actionToggle_Speech_to_Text,
            val,
            SignalCode.STT_LOAD_SIGNAL,
            SignalCode.STT_UNLOAD_SIGNAL,
            "stt_enabled"
        )
        QApplication.processEvents()
        self.update_application_settings("stt_enabled", val)
        if not val:
            self.emit_signal(SignalCode.STT_UNLOAD_SIGNAL)
        else:
            self.emit_signal(SignalCode.STT_LOAD_SIGNAL)

    @Slot()
    def action_stats_triggered(self):
        widget = StatsWidget()
        # display in a window
        widget.show()

    """
    End slot functions
    """
    def download_url(self, url, save_path):
        response = requests.get(url)
        soup = BeautifulSoup(response.content, 'html.parser')
        title = soup.title.string if soup.title else url
        # Truncate title to 10 words
        title_words = title.split()[:10]
        filename = "_".join(title_words) + '.html'
        # Replace any characters in filename that are not alphanumerics, underscores, or hyphens
        filename = re.sub(r'[^\w\-_]', '_', filename)
        save_path = os.path.join(save_path, filename)
        with open(save_path, 'wb') as file:
            file.write(response.content)
        return filename

    def download_pdf(self, url, save_path):
        response = requests.get(url)
        filename = url.split('/')[-1]  # Get the filename from the URL
        save_path = os.path.join(save_path, filename)
        with open(save_path, 'wb') as file:
            file.write(response.content)
        return filename

    def on_navigate_to_url(self, data: dict = None):
        url, ok = QInputDialog.getText(self, 'Browse Web', 'Enter your URL:')
        if ok:
            try:
                result = urllib.parse.urlparse(url)
                is_url = all([result.scheme, result.netloc])
            except ValueError:
                is_url = False

            # If the input is a URL, download it
            if is_url:
                if url.lower().endswith('.pdf'):
                    # Handle PDF file
                    filepath = os.path.expanduser(self.path_settings.pdf_path)
                    filename = self.download_pdf(url, filepath)
                else:
                    # Handle URL
                    filepath = os.path.expanduser(self.path_settings.webpages_path)
                    filename = self.download_url(url, filepath)
            elif os.path.isfile(url):
                filepath = os.path.dirname(url)
                filename = os.path.basename(url)
            else:
                self.logger.error(f"Invalid URL or file path")
                return

            # Update target files to use only the file that was downloaded or navigated to
            # and update the index.
            self.update_chatbot("target_files", [os.path.join(filepath, filename)])
            self.emit_signal(SignalCode.RAG_RELOAD_INDEX_SIGNAL, {
                "target_files": self.chatbot.target_files
            })
            self.emit_signal(
                SignalCode.LLM_TEXT_GENERATE_REQUEST_SIGNAL,
                {
                    "llm_request": True,
                    "request_data": {
                        "action": LLMAction.RAG,
                        "prompt": "Summarize the text and provide a synopsis of the content. Be concise and informative.",
                    }
                }
            )

    def on_describe_image_signal(self, data):
        image = data["image"]
        callback = data["callback"]
        self.generator_tab_widget.ui.ai_tab_widget.describe_image(
            image=image,
            callback=callback
        )

    def show_layers(self):
        self.emit_signal(SignalCode.LAYERS_SHOW_SIGNAL)

    def register_signals(self):
        self.logger.debug("Connecting signals")
        for item in (
            (SignalCode.SD_SAVE_PROMPT_SIGNAL, self.on_save_stablediffusion_prompt_signal),
            (SignalCode.QUIT_APPLICATION, self.action_quit_triggered),
            (SignalCode.SD_NSFW_CONTENT_DETECTED_SIGNAL, self.on_nsfw_content_detected_signal),
            (SignalCode.ENABLE_BRUSH_TOOL_SIGNAL, lambda _message: self.action_toggle_brush(True)),
            (SignalCode.ENABLE_ERASER_TOOL_SIGNAL, lambda _message: self.action_toggle_eraser(True)),
            (SignalCode.ENABLE_SELECTION_TOOL_SIGNAL, lambda _message: self.action_toggle_select(True)),
            (SignalCode.ENABLE_MOVE_TOOL_SIGNAL, lambda _message: self.action_toggle_active_grid_area(True)),
            (SignalCode.BASH_EXECUTE_SIGNAL, self.on_bash_execute_signal),
            (SignalCode.WRITE_FILE, self.on_write_file_signal),
            (SignalCode.TOGGLE_FULLSCREEN_SIGNAL, self.on_toggle_fullscreen_signal),
            (SignalCode.TOGGLE_TTS_SIGNAL, self.on_toggle_tts),
            (SignalCode.TOGGLE_SD_SIGNAL, self.on_toggle_sd),
            (SignalCode.TOGGLE_LLM_SIGNAL, self.on_toggle_llm),
            (SignalCode.UNLOAD_NON_SD_MODELS, self.on_unload_non_sd_models),
            (SignalCode.LOAD_NON_SD_MODELS, self.on_load_non_sd_models),
            (SignalCode.APPLICATION_RESET_SETTINGS_SIGNAL, self.action_reset_settings),
            (SignalCode.APPLICATION_RESET_PATHS_SIGNAL, self.on_reset_paths_signal),
            (SignalCode.MODEL_STATUS_CHANGED_SIGNAL, self.on_model_status_changed_signal),
            (SignalCode.KEYBOARD_SHORTCUTS_UPDATED, self.on_keyboard_shortcuts_updated),
            (SignalCode.HISTORY_UPDATED, self.on_history_updated),
            (SignalCode.REFRESH_STYLESHEET_SIGNAL, self.on_theme_changed_signal),
            (SignalCode.AI_MODELS_SAVE_OR_UPDATE_SIGNAL, self.on_ai_models_save_or_update_signal),
            (SignalCode.NAVIGATE_TO_URL, self.on_navigate_to_url),
        ):
            self.register(item[0], item[1])

    def on_reset_paths_signal(self):
        self.reset_path_settings()

    def restart(self):
        # Save the current state
        self.save_state()

        # Close the main window
        self.close()

        # Start a new instance of the application
        QProcess.startDetached(sys.executable, sys.argv)

    def on_write_file_signal(self, data: dict):
        """
        Writes data to a file.
        :param data: dict
        :return: None
        """
        args = data["args"]
        if len(args) == 1:
            message = args[0]
        else:
            message = args[1]
        with open("output.txt", "w") as f:
            f.write(message)

    def on_bash_execute_signal(self, data: dict) -> str:
        """
        Takes a message from the LLM and strips bash commands from it.
        Passes bash command to the bash_execute function.
        :param data: dict
        :return:
        """
        args = data["args"]
        return bash_execute(args[0])

    def on_theme_changed_signal(self):
        self.set_stylesheet()

    def initialize_ui(self):
        self.logger.debug("Loading UI")
        self.ui.setupUi(self)
        self.set_stylesheet()
        self.restore_state()
        self.status_widget = StatusWidget()
        self.statusBar().addPermanentWidget(self.status_widget)
        self.emit_signal(SignalCode.APPLICATION_CLEAR_STATUS_MESSAGE_SIGNAL)
        self.initialize_widget_elements()
        self.ui.actionUndo.setEnabled(False)
        self.ui.actionRedo.setEnabled(False)
        self.emit_signal(SignalCode.APPLICATION_MAIN_WINDOW_LOADED_SIGNAL, {"main_window": self})

    def initialize_widget_elements(self):
        for item in (
            (self.ui.actionToggle_LLM, self.application_settings.llm_enabled),
            (self.ui.actionToggle_Text_to_Speech, self.application_settings.tts_enabled),
            (self.ui.actionToggle_Speech_to_Text, self.application_settings.stt_enabled),
            (self.ui.actionToggle_Stable_Diffusion, self.application_settings.sd_enabled),
            (self.ui.actionToggle_Controlnet, self.application_settings.controlnet_enabled),
        ):
            item[0].blockSignals(True)
            item[0].setChecked(item[1] or False)
            item[0].blockSignals(False)
        self.initialized = True

    def layer_opacity_changed(self, attr_name, value=None, widget=None):
        self.emit_signal(SignalCode.LAYER_OPACITY_CHANGED_SIGNAL, value)

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        for v in self.shortcut_keys:
            if v.key == event.key():
                for signal in SignalCode:
                    if signal.value == v.signal:
                        self.emit_signal(signal)
                        break

    def key_text(self, key_name):
        for shortcutkey in self.shortcut_keys:
            if shortcutkey.name == key_name:
                return shortcutkey.text
        return ""

    def on_save_stablediffusion_prompt_signal(self, data: dict):
        self.create_saved_prompt({
            'prompt': data["prompt"],
            'negative_prompt': data["negative_prompt"],
            'secondary_prompt': data["secondary_prompt"],
            'secondary_negative_prompt': data["secondary_negative_prompt"],
        })

    def set_path_settings(self, key, val):
        self.update_path_settings(key, val)


    def action_quit_triggered(self):
        QApplication.quit()
        self.close()

    def on_nsfw_content_detected_signal(self):
        # display message in status
        self.emit_signal(
            SignalCode.APPLICATION_STATUS_ERROR_SIGNAL,
            NSFW_CONTENT_DETECTED_MESSAGE
        )

    def closeEvent(self, event) -> None:
        self.logger.debug("Quitting")
        self.save_state()
        self.emit_signal(SignalCode.QUIT_APPLICATION)


    def show_settings_path(self, name, default_path=None):
        path = getattr(self.path_settings, name)
        show_path(default_path if default_path and path == "" else path)

    def set_icons(self, icon_name, widget_name, theme):
        if not self.initialized:
            return

        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(f":/icons/{theme}/{icon_name}.svg"),
            QtGui.QIcon.Mode.Normal,
            QtGui.QIcon.State.Off)
        getattr(self.ui, widget_name).setIcon(icon)
        self.update()


    def toggle_nsfw_filter(self):
        self.set_nsfw_filter_tooltip()

    def set_nsfw_filter_tooltip(self):
        self.ui.actionSafety_Checker.setToolTip(
            f"Click to {'enable' if not self.application_settings.nsfw_filter else 'disable'} NSFW filter"
        )

    def on_toggle_fullscreen_signal(self):
        if self.isFullScreen():
            self.showNormal()
        else:
            self.showFullScreen()

    def on_unload_non_sd_models(self, data:dict=None):
        self._llm_generate_worker.on_llm_on_unload_signal()
        self._tts_generator_worker.unload()
        self._stt_audio_processor_worker.unload()
        callback = data.get("callback", None)
        if callback:
            callback(data)

    def on_load_non_sd_models(self, data:dict=None):
        self._llm_generate_worker.load()
        if self.application_settings.tts_enabled:
            self._tts_generator_worker.load()
        if self.application_settings.stt_enabled:
            self._stt_audio_processor_worker.load()
        callback = data.get("callback", None)
        if callback:
            callback(data)

    def on_toggle_llm(self, data:dict=None, val=None):
        if val is None:
            val = not self.application_settings.llm_enabled
        self._update_action_button(
            ModelType.LLM,
            self.ui.actionToggle_LLM,
            val,
            SignalCode.LLM_LOAD_SIGNAL,
            SignalCode.LLM_UNLOAD_SIGNAL,
            "llm_enabled",
            data
        )

    def on_toggle_sd(self, data:dict=None, val=None):
        if val is None:
            val = not self.application_settings.sd_enabled
        self._update_action_button(
            ModelType.SD,
            self.ui.actionToggle_Stable_Diffusion,
            val,
            SignalCode.SD_LOAD_SIGNAL,
            SignalCode.SD_UNLOAD_SIGNAL,
            "sd_enabled",
            data
        )

    def on_toggle_tts(self, data:dict=None, val=None):
        if val is None:
            val = not self.application_settings.sd_enabled
        self._update_action_button(
            ModelType.TTS,
            self.ui.actionToggle_Text_to_Speech,
            val,
            SignalCode.TTS_ENABLE_SIGNAL,
            SignalCode.TTS_DISABLE_SIGNAL,
            "tts_enabled",
            data
        )

    def _update_action_button(
        self,
        model_type,
        element,
        val:bool,
        load_signal: SignalCode,
        unload_signal: SignalCode,
        application_setting:str=None,
        data:dict=None
    ):
        if self._model_status[model_type] is ModelStatus.LOADING:
            val = not val
        element.blockSignals(True)
        element.setChecked(val)
        element.blockSignals(False)
        QApplication.processEvents()
        if application_setting:
            self.update_application_settings(application_setting, val)
        if self._model_status[model_type] is not ModelStatus.LOADING:
            if val:
                self.emit_signal(load_signal, data)
            else:
                self.emit_signal(unload_signal, data)

    def save_state(self):
        if self.quitting:
            return
        self.quitting = True
        self.logger.debug("Saving window state")

        self.save_window_settings("is_maximized", self.isMaximized())
        self.save_window_settings("is_fullscreen", self.isFullScreen())
        self.save_window_settings("llm_splitter", self.ui.tool_tab_widget.ui.llm_splitter.saveState())
        self.save_window_settings("content_splitter", self.ui.content_splitter.saveState())
        self.save_window_settings("generator_form_splitter", self.ui.generator_widget.ui.generator_form_splitter.saveState())
        self.save_window_settings("tool_tab_widget_index", self.ui.tool_tab_widget.ui.tool_tab_widget_container.currentIndex())
        self.save_window_settings("grid_settings_splitter", self.ui.tool_tab_widget.ui.grid_settings_splitter.saveState())
        self.save_window_settings("width", self.width())
        self.save_window_settings("height", self.height())
        self.save_window_settings("x_pos", self.pos().x())
        self.save_window_settings("y_pos", self.pos().y())
        self.save_window_settings("mode_tab_widget_index", self.ui.generator_widget.ui.generator_form_tabs.currentIndex())

    def restore_state(self):
        self.logger.debug("Restoring state")
        if self.window_settings is not None and self.window_settings.is_maximized:
            self.showMaximized()
        elif self.window_settings is not None and self.window_settings.is_fullscreen:
            self.showFullScreen()
        else:
            self.showNormal()

        self.ui.actionToggle_Grid.blockSignals(True)
        self.ui.actionToggle_Grid.setChecked(self.grid_settings.show_grid)
        self.ui.actionToggle_Grid.blockSignals(False)

        first_run = False
        splitters = [
            ("content_splitter", self.ui.content_splitter),
            ("llm_splitter", self.ui.tool_tab_widget.ui.llm_splitter),
            ("generator_form_splitter", self.ui.generator_widget.ui.generator_form_splitter),
            ("grid_settings_splitter", self.ui.tool_tab_widget.ui.grid_settings_splitter),
        ]
        for splitter_name, splitter in splitters:
            splitter_state = getattr(self.window_settings, splitter_name)
            if splitter_state is not None:
                splitter.blockSignals(True)
                splitter.restoreState(splitter_state)
                splitter.blockSignals(False)
            elif splitter_name == "content_splitter":
                first_run = True
                splitter.setSizes([self.width() - 200, 512, 200])

        self.setMinimumSize(100, 100)  # Set a reasonable minimum size

        width = int(self.window_settings.width)
        height = int(self.window_settings.height)
        if first_run:
            screen_geometry = QGuiApplication.primaryScreen().geometry()
            x_pos = (screen_geometry.width() - width) // 2
            y_pos = (screen_geometry.height() - height) // 2
        else:
            x_pos = int(self.window_settings.x_pos)
            y_pos = int(self.window_settings.y_pos)
        self.ui.generator_widget.ui.generator_form_tabs.setCurrentIndex(
            int(self.window_settings.mode_tab_widget_index)
        )

        self.ui.tool_tab_widget.ui.tool_tab_widget_container.setCurrentIndex(
            int(self.window_settings.tool_tab_widget_index)
        )

        self.resize(width, height)
        self.move(x_pos, y_pos)
        self.raise_()

    ##### End window properties #####
    #################################

    ###### Window handlers ######
    def cell_size_changed(self, val):
        self.update_grid_settings("cell_size", val)

    def line_width_changed(self, val):
        self.update_grid_settings("line_width", val)

    def line_color_changed(self, val):
        self.update_grid_settings("line_color", val)

    def snap_to_grid_changed(self, val):
        self.update_grid_settings("snap_to_grid", val)

    def canvas_color_changed(self, val):
        self.update_grid_settings("canvas_color", val)

    def action_ai_toggled(self, val):
        self.update_application_settings("ai_mode", val)

    def action_toggle_grid(self, val):
        self.update_grid_settings("show_grid", val)

    def show_nsfw_warning_popup(self):
        if self.application_settings.show_nsfw_warning:
            """
            Display a popup window which asks the user if they are sure they want to disable the NSFW filter
            along with a checkbox that allows the user to disable the warning in the future.
            """
            msg_box = QMessageBox()
            msg_box.setWindowTitle("Disable Safety Checker Warning")
            msg_box.setText(
                (
                    "WARNING\n\n"
                    "You are attempting to disable the safety checker (NSFW filter).\n"
                    "It is strongly recommended that you keep this enabled at all times.\n"
                    "The Safety Checker prevents potentially harmful content from being displayed.\n"
                    "Only disable it if you are sure the Image model you are using is not capable of generating harmful content.\n"
                    "Disabling the safety checker is intended as a last resort for continual false positives and as a research feature.\n"
                    "\n\n"
                    "Are you sure you want to disable the filter?"
                )
            )
            msg_box.setStandardButtons(QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
            msg_box.setDefaultButton(QMessageBox.StandardButton.No)

            # Create a QCheckBox
            checkbox = QCheckBox("Do not show this warning again")
            # Add the checkbox to the message box
            msg_box.setCheckBox(checkbox)

            result = msg_box.exec()

            if result == QMessageBox.StandardButton.Yes:
                self._disable_nsfw_filter(not checkbox.isChecked())

            self.ui.actionSafety_Checker.blockSignals(True)
            self.ui.actionSafety_Checker.setChecked(self.application_settings.nsfw_filter)
            self.ui.actionSafety_Checker.blockSignals(False)
        else:
            self._disable_nsfw_filter()

    def _disable_nsfw_filter(self, show_nsfw_warning=None):
        """
        Do not call this function directly.
        :return:
        """
        # User confirmed to disable the NSFW filter
        # Update the settings accordingly
        self.update_application_settings("nsfw_filter", False)
        # Update the show_nsfw_warning setting based on the checkbox state
        if show_nsfw_warning is not None:
            self.update_application_settings("show_nsfw_warning", show_nsfw_warning)
        self.toggle_nsfw_filter()
        self.emit_signal(SignalCode.SAFETY_CHECKER_UNLOAD_SIGNAL)

    ###### End window handlers ######

    def show_update_message(self):
        self.emit_signal(
            SignalCode.APPLICATION_STATUS_INFO_SIGNAL,
            f"New version available: {self.latest_version}"
        )

    def show_update_popup(self):
        self.update_popup = UpdateWindow()

    def show_setup_wizard(self):
        AppInstaller(close_on_cancel=False)

    def showEvent(self, event):
        super().showEvent(event)
        self.logger.debug("showEvent called, initializing window")
        self._initialize_window()
        self._initialize_default_buttons()
        self._initialize_filter_actions()
        self.initialized = True
        for icon_data in self.icons:
            self.set_icons(
                icon_data[0],
                icon_data[1],
                "dark" if self.application_settings.dark_mode_enabled else "light"
            )
        self.logger.debug("Showing window")
        self._set_keyboard_shortcuts()

    def on_keyboard_shortcuts_updated(self):
        self._set_keyboard_shortcuts()

    def on_history_updated(self, data):
        self.ui.actionUndo.setEnabled(data["undo"] != 0)
        self.ui.actionRedo.setEnabled(data["redo"] != 0)

    def _set_keyboard_shortcuts(self):
        
        quit_key = self.session.query(ShortcutKeys).filter_by(display_name="Quit").first()
        brush_key = self.session.query(ShortcutKeys).filter_by(display_name="Brush").first()
        eraser_key = self.session.query(ShortcutKeys).filter_by(display_name="Eraser").first()
        move_tool_key = self.session.query(ShortcutKeys).filter_by(display_name="Move Tool").first()
        select_tool_key = self.session.query(ShortcutKeys).filter_by(display_name="Select Tool").first()

        if quit_key is not None:
            key_sequence = QKeySequence(quit_key.key | quit_key.modifiers)
            self.ui.actionQuit.setShortcut(key_sequence)
            self.ui.actionQuit.setToolTip(f"{quit_key.display_name} ({quit_key.text})")

        if brush_key is not None:
            key_sequence = QKeySequence(brush_key.key | brush_key.modifiers)
            self.ui.actionToggle_Brush.setShortcut(key_sequence)
            self.ui.actionToggle_Brush.setToolTip(f"{brush_key.display_name} ({brush_key.text})")

        if eraser_key is not None:
            key_sequence = QKeySequence(eraser_key.key | eraser_key.modifiers)
            self.ui.actionToggle_Eraser.setShortcut(key_sequence)
            self.ui.actionToggle_Eraser.setToolTip(f"{eraser_key.display_name} ({eraser_key.text})")

        if move_tool_key is not None:
            key_sequence = QKeySequence(move_tool_key.key | move_tool_key.modifiers)
            self.ui.actionToggle_Active_Grid_Area.setShortcut(key_sequence)
            self.ui.actionToggle_Active_Grid_Area.setToolTip(f"{move_tool_key.display_name} ({move_tool_key.text})")

        if select_tool_key is not None:
            key_sequence = QKeySequence(select_tool_key.key | select_tool_key.modifiers)
            self.ui.actionToggle_Selection.setShortcut(key_sequence)
            self.ui.actionToggle_Selection.setToolTip(f"{select_tool_key.display_name} ({select_tool_key.text})")

        

    def _initialize_workers(self):
        self.logger.debug("Initializing worker manager")
        self._mask_generator_worker = create_worker(MaskGeneratorWorker)
        self._sd_worker = create_worker(SDWorker)
        self._stt_audio_capture_worker = create_worker(AudioCaptureWorker)
        self._stt_audio_processor_worker = create_worker(AudioProcessorWorker)
        self._tts_generator_worker = create_worker(TTSGeneratorWorker)
        self._tts_vocalizer_worker = create_worker(TTSVocalizerWorker)
        self._llm_generate_worker = create_worker(LLMGenerateWorker)

    def _initialize_filter_actions(self):
        # add more filters:
        
        image_filters = self.session.query(ImageFilter).all()
        for image_filter in image_filters:
            action = self.ui.menuFilters.addAction(image_filter.display_name)
            action.triggered.connect(partial(self.display_filter_window, image_filter))
        

    def display_filter_window(self, image_filter):
        FilterWindow(image_filter.id)

    def _initialize_default_buttons(self):
        show_grid = self.grid_settings.show_grid
        current_tool = self.current_tool

        set_widget_state(self.ui.actionToggle_Active_Grid_Area, current_tool is CanvasToolName.ACTIVE_GRID_AREA)
        set_widget_state(self.ui.actionToggle_Brush, current_tool is CanvasToolName.BRUSH)
        set_widget_state(self.ui.actionToggle_Eraser, current_tool is CanvasToolName.ERASER)
        set_widget_state(self.ui.actionToggle_Grid, show_grid is True)
        set_widget_state(self.ui.actionMask_toggle, self.drawing_pad_settings.mask_layer_enabled is True)

        self.ui.actionSafety_Checker.blockSignals(True)
        self.ui.actionSafety_Checker.setChecked(self.application_settings.nsfw_filter)
        self.ui.actionSafety_Checker.blockSignals(False)

    @staticmethod
    def __toggle_button(ui_element, state):
        ui_element.blockSignals(True)
        ui_element.setChecked(state)
        ui_element.blockSignals(False)

    def toggle_tool(self, tool: CanvasToolName, active: bool):
        self.initialize_widget_elements()
        if not active:
            tool = CanvasToolName.NONE
        else:
            if tool is CanvasToolName.BRUSH:
                self.__toggle_button(self.ui.actionToggle_Brush, True)
                self.__toggle_button(self.ui.actionToggle_Eraser, False)
                self.__toggle_button(self.ui.actionToggle_Selection, False)
                self.__toggle_button(self.ui.actionToggle_Active_Grid_Area, False)
            elif tool is CanvasToolName.ERASER:
                self.__toggle_button(self.ui.actionToggle_Brush, False)
                self.__toggle_button(self.ui.actionToggle_Eraser, True)
                self.__toggle_button(self.ui.actionToggle_Selection, False)
                self.__toggle_button(self.ui.actionToggle_Active_Grid_Area, False)
            elif tool is CanvasToolName.SELECTION:
                self.__toggle_button(self.ui.actionToggle_Brush, False)
                self.__toggle_button(self.ui.actionToggle_Eraser, False)
                self.__toggle_button(self.ui.actionToggle_Selection, True)
                self.__toggle_button(self.ui.actionToggle_Active_Grid_Area, False)
            elif tool is CanvasToolName.ACTIVE_GRID_AREA:
                self.__toggle_button(self.ui.actionToggle_Brush, False)
                self.__toggle_button(self.ui.actionToggle_Eraser, False)
                self.__toggle_button(self.ui.actionToggle_Selection, False)
                self.__toggle_button(self.ui.actionToggle_Active_Grid_Area, True)
        self.update_application_settings("current_tool", tool.value)
        self.emit_signal(SignalCode.APPLICATION_TOOL_CHANGED_SIGNAL, {
            "tool": tool
        })
        self.emit_signal(SignalCode.CANVAS_UPDATE_CURSOR)

    def _initialize_window(self):
        self.center()
        self.set_window_title()

    def center(self):
        availableGeometry = QGuiApplication.primaryScreen().availableGeometry()
        frameGeometry = self.frameGeometry()
        frameGeometry.moveCenter(availableGeometry.center())
        self.move(frameGeometry.topLeft())

    def set_window_title(self):
        """
        Overrides base method to set the window title
        :return:
        """
        self.setWindowTitle(self._window_title)

    def new_document(self):
        self.is_saved = False
        self.set_window_title()
        self.current_filter = None

    def handle_unknown(self, message):
        self.logger.error(f"Unknown message code: {message}")

    def clear_all_prompts(self):
        self.prompt = ""
        self.negative_prompt = ""
        self.generator_tab_widget.clear_prompts()

    def new_batch(self, index, image, data):
        self.generator_tab_widget.new_batch(index, image, data)

    def on_model_status_changed_signal(self, data):
        model = data["model"]
        status = data["status"]
        if self._model_status[model] is status:
            return
        self._model_status[model] = status
        if model is ModelType.SD:
            self.ui.actionToggle_Stable_Diffusion.setDisabled(status is ModelStatus.LOADING)
        elif model is ModelType.CONTROLNET:
            self.ui.actionToggle_Controlnet.setDisabled(status is ModelStatus.LOADING)
        elif model is ModelType.LLM:
            self.ui.actionToggle_LLM.setDisabled(status is ModelStatus.LOADING)
        elif model is ModelType.TTS:
            self.ui.actionToggle_Text_to_Speech.setDisabled(status is ModelStatus.LOADING)
        elif model is ModelType.STT:
            self.ui.actionToggle_Speech_to_Text.setDisabled(status is ModelStatus.LOADING)
        self.initialize_widget_elements()
        QApplication.processEvents()

    def _generate_drawingpad_mask(self):
        width = self.active_grid_settings.width
        height = self.active_grid_settings.height
        img = Image.new("RGB", (width, height), (0, 0, 0))
        base64_image = convert_image_to_binary(img)
        
        drawing_pad_settings = self.session.query(DrawingPadSettings).first()
        drawing_pad_settings.mask = base64_image
        self.session.commit()
        



def prepare_llm_generate_kwargs(settings):
    min_val = 0.0001
    length_penalty = settings.length_penalty / 1000.0
    repetition_penalty = settings.repetition_penalty / 100.0
    top_p = settings.top_p / 1000.0
    temperature = settings.temperature / 10000.0

    if length_penalty < min_val:
        length_penalty = min_val

    if repetition_penalty < min_val:
        repetition_penalty = min_val

    if top_p < min_val:
        top_p = min_val

    if temperature < min_val:
        temperature = min_val

    return dict(
        # length_penalty=length_penalty,
        # repetition_penalty=repetition_penalty,
        do_sample=settings.do_sample,
        early_stopping=settings.early_stopping,
        eta_cutoff=settings.eta_cutoff,
        max_new_tokens=settings.max_new_tokens,
        min_length=settings.min_length,
        # no_repeat_ngram_size=data["ngram_size"],
        num_return_sequences=1,
        temperature=temperature,
        top_k=settings.top_k,
        top_p=top_p,
        # use_cache=data["use_cache"],
        # num_beams=data["num_beams"],
    )



import threading

from airunner.handlers.llm.causal_lm_transformer_base_handler import CausalLMTransformerBaseHandler
from airunner.enums import SignalCode
from airunner.workers.worker import Worker


class LLMGenerateWorker(Worker):
    def __init__(self, agent_options=None):
        self.llm = None
        self.agent_options = agent_options
        super().__init__()
        for signal in (
            (SignalCode.LLM_REQUEST_WORKER_RESPONSE_SIGNAL, self.on_llm_request_worker_response_signal),
            (SignalCode.LLM_UNLOAD_SIGNAL, self.on_llm_on_unload_signal),
            (SignalCode.LLM_LOAD_SIGNAL, self.on_llm_load_model_signal),
            (SignalCode.LLM_CLEAR_HISTORY_SIGNAL, self.on_llm_clear_history_signal),
            (SignalCode.LLM_TEXT_GENERATE_REQUEST_SIGNAL, self.on_llm_request_signal),
            (SignalCode.RAG_RELOAD_INDEX_SIGNAL, self.on_llm_reload_rag_index_signal),
            (SignalCode.ADD_CHATBOT_MESSAGE_SIGNAL, self.on_llm_add_chatbot_response_to_history),
            (SignalCode.LOAD_CONVERSATION, self.on_llm_load_conversation),
            (SignalCode.INTERRUPT_PROCESS_SIGNAL, self.llm_on_interrupt_process_signal),
            (SignalCode.QUIT_APPLICATION, self.on_quit_application_signal),
        ):
            self.register(signal[0], signal[1])
        self._llm_thread = None

    def on_quit_application_signal(self):
        self.logger.debug("Quitting LLM")
        if self.llm:
            self.llm.unload()
        if self._llm_thread is not None:
            self._llm_thread.join()

    def on_llm_request_worker_response_signal(self, message: dict):
        self.add_to_queue(message)

    def on_llm_on_unload_signal(self, data=None):
        data = data or {}
        self.logger.debug("Unloading LLM")
        self.llm.unload()
        callback = data.get("callback", None)
        if callback:
            callback(data)

    def on_llm_load_model_signal(self, data):
        self._load_llm_thread(data)

    def on_llm_clear_history_signal(self):
        if self.llm:
            self.llm.clear_history()

    def on_llm_request_signal(self, message: dict):
        self.add_to_queue(message)

    def llm_on_interrupt_process_signal(self):
        if self.llm:
            self.llm.do_interrupt()

    def on_llm_reload_rag_index_signal(self):
        if self.llm:
            self.llm.reload_rag()

    def on_llm_add_chatbot_response_to_history(self, message):
        self.llm.add_chatbot_response_to_history(message)

    def on_llm_load_conversation(self, message):
        try:
            self.llm.load_conversation(message)
        except Exception as e:
            self.logger.error(f"Error in on_load_conversation: {e}")

    def start_worker_thread(self):
        if self.application_settings.llm_enabled:
            self._load_llm_thread()

    def handle_message(self, message):
        self.llm.handle_request(message)

    def _load_llm_thread(self, data=None):
        self._llm_thread = threading.Thread(target=self._load_llm, args=(data,))
        self._llm_thread.start()

    def load(self):
        self._load_llm()

    def _load_llm(self, data=None):
        data = data or {}
        if self.llm is None:
            self.llm = CausalLMTransformerBaseHandler(agent_options=self.agent_options)

        self.llm.load()

        callback = data.get("callback", None)
        if callback:
            callback(data)
